# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from pyqrllib.pyqrllib import shake256

from qrl.core.misc import logger

logger.initialize_default()


class RNG(object):
    """
    A naive implementaiton of RNG
    WARNING: Only for unit testing.
    """
    @staticmethod
    def generate(prf_seed: bytes, seq: int) -> bytes:
        prf = prf_seed
        for i in range(seq):
            prf = shake256(32, prf)

        return bytes(prf)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

import qrl.generated.qrlmining_pb2 as qrlmining__pb2


class MiningAPIStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.GetBlockMiningCompatible = channel.unary_unary(
        '/qrl.MiningAPI/GetBlockMiningCompatible',
        request_serializer=qrlmining__pb2.GetBlockMiningCompatibleReq.SerializeToString,
        response_deserializer=qrlmining__pb2.GetBlockMiningCompatibleResp.FromString,
        )
    self.GetLastBlockHeader = channel.unary_unary(
        '/qrl.MiningAPI/GetLastBlockHeader',
        request_serializer=qrlmining__pb2.GetLastBlockHeaderReq.SerializeToString,
        response_deserializer=qrlmining__pb2.GetLastBlockHeaderResp.FromString,
        )
    self.GetBlockToMine = channel.unary_unary(
        '/qrl.MiningAPI/GetBlockToMine',
        request_serializer=qrlmining__pb2.GetBlockToMineReq.SerializeToString,
        response_deserializer=qrlmining__pb2.GetBlockToMineResp.FromString,
        )
    self.SubmitMinedBlock = channel.unary_unary(
        '/qrl.MiningAPI/SubmitMinedBlock',
        request_serializer=qrlmining__pb2.SubmitMinedBlockReq.SerializeToString,
        response_deserializer=qrlmining__pb2.SubmitMinedBlockResp.FromString,
        )


class MiningAPIServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def GetBlockMiningCompatible(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetLastBlockHeader(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBlockToMine(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SubmitMinedBlock(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_MiningAPIServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'GetBlockMiningCompatible': grpc.unary_unary_rpc_method_handler(
          servicer.GetBlockMiningCompatible,
          request_deserializer=qrlmining__pb2.GetBlockMiningCompatibleReq.FromString,
          response_serializer=qrlmining__pb2.GetBlockMiningCompatibleResp.SerializeToString,
      ),
      'GetLastBlockHeader': grpc.unary_unary_rpc_method_handler(
          servicer.GetLastBlockHeader,
          request_deserializer=qrlmining__pb2.GetLastBlockHeaderReq.FromString,
          response_serializer=qrlmining__pb2.GetLastBlockHeaderResp.SerializeToString,
      ),
      'GetBlockToMine': grpc.unary_unary_rpc_method_handler(
          servicer.GetBlockToMine,
          request_deserializer=qrlmining__pb2.GetBlockToMineReq.FromString,
          response_serializer=qrlmining__pb2.GetBlockToMineResp.SerializeToString,
      ),
      'SubmitMinedBlock': grpc.unary_unary_rpc_method_handler(
          servicer.SubmitMinedBlock,
          request_deserializer=qrlmining__pb2.SubmitMinedBlockReq.FromString,
          response_serializer=qrlmining__pb2.SubmitMinedBlockResp.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'qrl.MiningAPI', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from qrl.core.txs.Transaction import Transaction
from qrl.generated import qrlstateinfo_pb2


class LastTransactions:
    def __init__(self, pbdata=None):
        self._data = pbdata
        if not pbdata:
            self._data = qrlstateinfo_pb2.LastTransactions()

    @property
    def tx_metadata(self):
        return self._data.tx_metadata

    def add(self, tx: Transaction, block_number: int, timestamp: int):
        tm = qrlstateinfo_pb2.TransactionMetadata(transaction=tx.pbdata,
                                                  block_number=block_number,
                                                  timestamp=timestamp)
        tmp = self._data.tx_metadata[::-1]
        tmp.append(tm)
        del self._data.tx_metadata[:]
        self._data.tx_metadata.extend(tmp[-20:][::-1])

    def serialize(self) -> str:
        return self._data.SerializeToString()

    @staticmethod
    def deserialize(data):
        pbdata = qrlstateinfo_pb2.LastTransactions()
        pbdata.ParseFromString(bytes(data))
        return LastTransactions(pbdata)



from ipaddress import IPv4Address

from qrl.core import config


class IPMetadata(object):
    def __init__(self, ip_str: str, port: int):
        self._ip = ip_str

        try:
            self._port = int(port)
        except ValueError:
            raise ValueError('Invalid Peer Port {}'.format(port))

        self.ip_address = IPv4Address(self._ip)

        if not (0 < self._port <= 65535):  # Validate port number
            raise ValueError('Invalid Peer Port {}'.format(self))

    def __repr__(self):
        return self.full_address

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        if isinstance(other, IPMetadata):
            return self._ip == other._ip and self._port == other._port
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def _validate(self):
        pass

    @property
    def full_address(self):
        return "{}:{}".format(self.ip, self.port)

    @property
    def ip(self):
        return self._ip

    @property
    def port(self):
        return self._port

    @property
    def is_global(self):
        return self.ip_address.is_global

    @classmethod
    def from_full_address(cls, full_address: str, check_global=False):
        parts = full_address.split(':')

        ip = parts[0]
        port = config.user.p2p_local_port
        if check_global:
            port = config.user.p2p_public_port

        if len(parts) > 2:
            raise ValueError('Invalid Peer address')

        if len(parts) == 2:
            try:
                port = int(parts[1])
            except ValueError as e:
                raise ValueError('Invalid Peer Port {} - {}'.format(full_address, str(e)))

        answer = cls(ip, port)
        if check_global:
            if not answer.is_global:  # Check for Global IP
                raise ValueError('Local Peer IP Found {}'.format(full_address))

        return answer

    @staticmethod
    def canonical_full_address(full_address: str, check_global=False):
        return IPMetadata.from_full_address(full_address, check_global).full_address



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from decimal import Decimal
from typing import Optional, List, Iterator, Tuple

from pyqrllib.pyqrllib import QRLHelper, bin2hstr
from twisted.internet import reactor

from qrl.core import config
from qrl.core.AddressState import AddressState
from qrl.core.Block import Block
from qrl.core.ChainManager import ChainManager
from qrl.core.ESyncState import ESyncState
from qrl.core.misc import ntp
from qrl.core.misc.logger import logger
from qrl.core.node import POW, SyncState
from qrl.core.p2p.p2pChainManager import P2PChainManager
from qrl.core.p2p.p2pPeerManager import P2PPeerManager
from qrl.core.p2p.p2pTxManagement import P2PTxManagement
from qrl.core.p2p.p2pfactory import P2PFactory
from qrl.core.txs.CoinBase import CoinBase
from qrl.core.txs.MessageTransaction import MessageTransaction
from qrl.core.txs.SlaveTransaction import SlaveTransaction
from qrl.core.txs.TokenTransaction import TokenTransaction
from qrl.core.txs.TransferTokenTransaction import TransferTokenTransaction
from qrl.core.txs.TransferTransaction import TransferTransaction
from qrl.generated import qrl_pb2


class QRLNode:
    def __init__(self, mining_address: bytes):
        self.start_time = ntp.getTime()
        self._sync_state = SyncState()

        self.peer_manager = P2PPeerManager()
        self.peer_manager.load_peer_addresses()
        self.peer_manager.register(P2PPeerManager.EventType.NO_PEERS, self.connect_peers)

        self.p2pchain_manager = P2PChainManager()

        self.tx_manager = P2PTxManagement()

        self._chain_manager = None  # FIXME: REMOVE. This is temporary
        self._p2pfactory = None  # FIXME: REMOVE. This is temporary

        self._pow = None

        self.mining_address = mining_address

        reactor.callLater(10, self.monitor_chain_state)

    ####################################################
    ####################################################
    ####################################################
    ####################################################

    @property
    def version(self):
        return config.dev.version

    @property
    def sync_state(self) -> SyncState:
        return self._sync_state

    @property
    def state(self):
        if self._p2pfactory is None:
            return ESyncState.unknown.value
        # FIXME
        return self._p2pfactory.sync_state.state.value

    @property
    def num_connections(self):
        if self._p2pfactory is None:
            return 0
        return self._p2pfactory.num_connections

    @property
    def num_known_peers(self):
        return len(self.peer_manager.known_peer_addresses)

    @property
    def uptime(self):
        return ntp.getTime() - self.start_time

    @property
    def block_height(self):
        return self._chain_manager.height

    @property
    def epoch(self):
        if not self._chain_manager.last_block:
            return 0
        return self._chain_manager.last_block.block_number // config.dev.blocks_per_epoch

    @property
    def uptime_network(self):
        block_one = self._chain_manager.get_block_by_number(1)
        network_uptime = 0
        if block_one:
            network_uptime = ntp.getTime() - block_one.timestamp
        return network_uptime

    @property
    def block_last_reward(self):
        if not self._chain_manager.last_block:
            return 0

        return self._chain_manager.last_block.block_reward

    @property
    def block_time_mean(self):
        block = self._chain_manager.last_block

        prev_block_metadata = self._chain_manager.get_block_metadata(block.prev_headerhash)
        if prev_block_metadata is None:
            return config.dev.mining_setpoint_blocktime

        movavg = self._chain_manager.get_measurement(block.timestamp,
                                                     block.prev_headerhash,
                                                     prev_block_metadata)
        return movavg

    @property
    def block_time_sd(self):
        # FIXME: Keep a moving var
        return 0

    @property
    def coin_supply(self):
        # FIXME: Keep a moving var
        return self._chain_manager.total_coin_supply

    @property
    def coin_supply_max(self):
        # FIXME: Keep a moving var
        return config.dev.max_coin_supply

    ####################################################
    ####################################################
    ####################################################
    ####################################################

    def get_peers_stat(self) -> list:
        return self.peer_manager.get_peers_stat()

    def connect_peers(self):
        self.peer_manager.connect_peers()

    ####################################################
    ####################################################
    ####################################################
    ####################################################

    def monitor_chain_state(self):
        self.peer_manager.monitor_chain_state()

        last_block = self._chain_manager.last_block
        block_metadata = self._chain_manager.get_block_metadata(last_block.headerhash)
        node_chain_state = qrl_pb2.NodeChainState(block_number=last_block.block_number,
                                                  header_hash=last_block.headerhash,
                                                  cumulative_difficulty=bytes(block_metadata.cumulative_difficulty),
                                                  version=config.dev.version,
                                                  timestamp=ntp.getTime())

        self.peer_manager.broadcast_chain_state(node_chain_state=node_chain_state)
        channel = self.peer_manager.get_better_difficulty(block_metadata.cumulative_difficulty)
        logger.debug('Got better difficulty %s', channel)
        if channel:
            logger.debug('Connection id >> %s', channel.peer)
            channel.send_get_headerhash_list(self._chain_manager.height)
        reactor.callLater(config.user.chain_state_broadcast_period, self.monitor_chain_state)

    # FIXME: REMOVE. This is temporary
    def set_chain_manager(self, chain_manager: ChainManager):
        self._chain_manager = chain_manager

    ####################################################
    ####################################################
    ####################################################
    ####################################################

    def start_pow(self, mining_thread_count):
        self._pow = POW(chain_manager=self._chain_manager,
                        p2p_factory=self._p2pfactory,
                        sync_state=self._sync_state,
                        time_provider=ntp,
                        mining_address=self.mining_address,
                        mining_thread_count=mining_thread_count)

        self._pow.start()

    def start_listening(self):
        self._p2pfactory = P2PFactory(chain_manager=self._chain_manager,
                                      sync_state=self.sync_state,
                                      qrl_node=self)  # FIXME: Try to avoid cyclic references

        self.peer_manager._p2pfactory = self._p2pfactory
        self._p2pfactory.start_listening()

    ####################################################
    ####################################################
    ####################################################
    ####################################################

    @staticmethod
    def validate_amount(amount_str: str) -> bool:
        # FIXME: Refactored code. Review Decimal usage all over the code
        Decimal(amount_str)
        return True

    ####################################################
    ####################################################
    ####################################################
    ####################################################

    @staticmethod
    def create_message_txn(message_hash: bytes,
                           fee: int,
                           xmss_pk: bytes,
                           master_addr: bytes):
        return MessageTransaction.create(message_hash=message_hash,
                                         fee=fee,
                                         xmss_pk=xmss_pk,
                                         master_addr=master_addr)

    @staticmethod
    def create_token_txn(symbol: bytes,
                         name: bytes,
                         owner: bytes,
                         decimals: int,
                         initial_balances,
                         fee: int,
                         xmss_pk: bytes,
                         master_addr: bytes):
        return TokenTransaction.create(symbol,
                                       name,
                                       owner,
                                       decimals,
                                       initial_balances,
                                       fee,
                                       xmss_pk,
                                       master_addr)

    @staticmethod
    def create_transfer_token_txn(addrs_to: list,
                                  token_txhash: bytes,
                                  amounts: list,
                                  fee: int,
                                  xmss_pk: bytes,
                                  master_addr: bytes):
        return TransferTokenTransaction.create(token_txhash,
                                               addrs_to,
                                               amounts,
                                               fee,
                                               xmss_pk,
                                               master_addr)

    def create_send_tx(self,
                       addrs_to: list,
                       amounts: list,
                       fee: int,
                       xmss_pk: bytes,
                       master_addr: bytes) -> TransferTransaction:
        addr_from = self.get_addr_from(xmss_pk, master_addr)
        balance = self._chain_manager.get_address_balance(addr_from)
        if sum(amounts) + fee > balance:
            raise ValueError("Not enough funds in the source address")

        return TransferTransaction.create(addrs_to=addrs_to,
                                          amounts=amounts,
                                          fee=fee,
                                          xmss_pk=xmss_pk,
                                          master_addr=master_addr)

    def create_slave_tx(self,
                        slave_pks: list,
                        access_types: list,
                        fee: int,
                        xmss_pk: bytes,
                        master_addr: bytes) -> SlaveTransaction:
        return SlaveTransaction.create(slave_pks=slave_pks,
                                       access_types=access_types,
                                       fee=fee,
                                       xmss_pk=xmss_pk,
                                       master_addr=master_addr)

    # FIXME: Rename this appropriately
    def submit_send_tx(self, tx) -> bool:
        if tx is None:
            raise ValueError("The transaction was empty")

        if self._chain_manager.tx_pool.is_full_pending_transaction_pool():
            raise ValueError("Pending Transaction Pool is full")

        return self._p2pfactory.add_unprocessed_txn(tx, ip=None)  # TODO (cyyber): Replace None with IP made API request

    @staticmethod
    def get_addr_from(xmss_pk, master_addr):
        if master_addr:
            return master_addr

        return bytes(QRLHelper.getAddress(xmss_pk))

    def get_address_is_used(self, address: bytes) -> bool:
        if not AddressState.address_is_valid(address):
            raise ValueError("Invalid Address")

        return self._chain_manager.get_address_is_used(address)

    def get_address_state(self, address: bytes) -> AddressState:
        if address != config.dev.coinbase_address and not AddressState.address_is_valid(address):
            raise ValueError("Invalid Address")

        address_state = self._chain_manager.get_address_state(address)

        return address_state

    def get_all_address_state(self) -> list:
        return self._chain_manager.get_all_address_state()

    def get_transactions_by_address(self, address: bytes):
        address_state = self._chain_manager.get_address_state(address)
        mini_transactions = []
        balance = 0
        for tx_hash in address_state.transaction_hashes:
            mini_transaction = qrl_pb2.MiniTransaction()
            mini_transaction.transaction_hash = bin2hstr(tx_hash)
            tx, _ = self._chain_manager.get_tx_metadata(tx_hash)
            amount = 0
            if tx.addr_from == address:
                amount -= tx.fee
            if isinstance(tx, TransferTransaction):
                if tx.addr_from == address:
                    amount -= tx.total_amount
                try:
                    for i in range(len(tx.addrs_to)):
                        if tx.addrs_to[i] == address:
                            amount += tx.amounts[i]
                except ValueError:
                    pass
            elif isinstance(tx, CoinBase):
                if tx.addr_to == address:
                    amount += tx.amount

            if amount < 0:
                mini_transaction.out = True
            mini_transaction.amount = abs(amount)
            mini_transactions.append(mini_transaction)
            balance += amount

        return mini_transactions, balance

    def get_transaction(self, query_hash: bytes):
        """
        This method returns an object that matches the query hash
        """
        # FIXME: At some point, all objects in DB will indexed by a hash
        # TODO: Search tx hash
        # FIXME: We dont need searches, etc.. getting a protobuf indexed by hash from DB should be enough
        # FIXME: This is just a workaround to provide functionality
        result = self._chain_manager.get_tx_metadata(query_hash)
        return result

    def get_block_header_hash_by_number(self, query_block_number: int):
        return self._chain_manager.get_block_header_hash_by_number(query_block_number)

    def get_unconfirmed_transaction(self, query_hash: bytes):
        result = self._chain_manager.get_unconfirmed_transaction(query_hash)
        return result

    def get_block_last(self) -> Optional[Block]:
        """
        This method returns an object that matches the query hash
        """
        return self._chain_manager.last_block

    def get_block_from_hash(self, query_hash: bytes) -> Optional[Block]:
        """
        This method returns an object that matches the query hash
        """
        return self._chain_manager.get_block(query_hash)

    def get_block_from_index(self, index: int) -> Block:
        """
        This method returns an object that matches the query hash
        """
        return self._chain_manager.get_block_by_number(index)

    def get_blockidx_from_txhash(self, transaction_hash):
        result = self._chain_manager.get_tx_metadata(transaction_hash)
        if result:
            return result[1]
        return None

    def get_latest_blocks(self, offset, count) -> List[Block]:
        answer = []
        end = self.block_height - offset
        start = max(0, end - count + 1)
        for blk_idx in range(start, end + 1):
            answer.append(self._chain_manager.get_block_by_number(blk_idx))

        return answer

    def get_latest_transactions(self, offset, count):
        answer = []
        skipped = 0
        for tx in self._chain_manager.get_last_transactions():
            if skipped >= offset:
                answer.append(tx)
                if len(answer) >= count:
                    break
            else:
                skipped += 1

        return answer

    def get_latest_transactions_unconfirmed(self, offset, count):
        answer = []
        skipped = 0
        for tx_set in self._chain_manager.tx_pool.transactions:
            if skipped >= offset:
                answer.append(tx_set[1])
                if len(answer) >= count:
                    break
            else:
                skipped += 1
        return answer

    def get_node_info(self) -> qrl_pb2.NodeInfo:
        info = qrl_pb2.NodeInfo()
        info.version = self.version
        info.state = self.state
        info.num_connections = self.num_connections
        info.num_known_peers = self.num_known_peers
        info.uptime = self.uptime
        info.block_height = self.block_height
        info.block_last_hash = self._chain_manager.last_block.headerhash
        info.network_id = config.user.genesis_prev_headerhash
        return info

    def get_block_timeseries(self, block_count) -> Iterator[qrl_pb2.BlockDataPoint]:
        result = []

        if self.block_height <= 0:
            return result

        block = self._chain_manager.last_block
        if block is None:
            return result

        headerhash_current = block.headerhash
        while len(result) < block_count:
            data_point = self._chain_manager.get_block_datapoint(headerhash_current)

            if data_point is None:
                break

            result.append(data_point)
            headerhash_current = data_point.header_hash_prev

        return reversed(result)

    def get_blockheader_and_metadata(self, block_number=0) -> Tuple:
        return self._chain_manager.get_blockheader_and_metadata(block_number)

    def get_block_to_mine(self, wallet_address) -> list:
        return self._chain_manager.get_block_to_mine(self._pow.miner, wallet_address)

    def submit_mined_block(self, blob) -> bool:
        return self._pow.miner.submit_mined_block(blob)



from unittest import TestCase

from mock import patch, Mock

from qrl.core.misc import logger
from qrl.core.AddressState import AddressState
from qrl.core.txs.MessageTransaction import MessageTransaction
from tests.misc.helper import get_alice_xmss, get_bob_xmss

logger.initialize_default()


@patch('qrl.core.txs.Transaction.logger')
class TestTransactionValidateSlave(TestCase):
    def setUp(self):
        self.alice = get_alice_xmss()
        self.params = {
            "message_hash": b'Test Message',
            "fee": 1,
            "xmss_pk": self.alice.pk
        }
        self.m_addr_state = Mock(autospec=AddressState, name='addr_state', balance=200)
        self.m_addr_from_pk_state = Mock(autospec=AddressState, name='addr_from_pk_state')

    def test_validate_slave_valid(self, m_logger):
        tx = MessageTransaction.create(**self.params)
        tx.sign(self.alice)
        result = tx.validate_slave(self.m_addr_state, self.m_addr_from_pk_state)
        self.assertTrue(result)

    def test_validate_slave_master_addr_same_as_signing_addr(self, m_logger):
        self.params["master_addr"] = self.alice.address
        tx = MessageTransaction.create(**self.params)
        tx.sign(self.alice)
        result = tx.validate_slave(self.m_addr_state, self.m_addr_from_pk_state)
        self.assertFalse(result)

    def test_validate_slave_signing_xmss_state_has_no_slave_permissions_in_state(self, m_logger):
        bob = get_bob_xmss()
        # Let's say Alice is Bob's master.
        self.params["master_addr"] = self.alice.address
        self.params["xmss_pk"] = bob.pk

        # We need to add extra data to the mock AddressState.
        self.m_addr_state.slave_pks_access_type = {}
        tx = MessageTransaction.create(**self.params)
        tx.sign(self.alice)
        result = tx.validate_slave(self.m_addr_state, self.m_addr_from_pk_state)
        self.assertFalse(result)

    def test_validate_slave_has_insufficient_permissions(self, m_logger):
        """
        Master's AddressState says the Slave has permission 0.
        But Slave's AddressState says the Slave is good for permission 2.
        Therefore the Slave does not have enough permissions.
        """
        bob = get_bob_xmss()
        # Let's say Alice is Bob's master.
        self.params["master_addr"] = self.alice.address
        self.params["xmss_pk"] = bob.pk

        tx = MessageTransaction.create(**self.params)
        tx.sign(self.alice)

        # The master's state says the slave can have these permissions.
        self.m_addr_state.slave_pks_access_type = {str(tx.PK): 0}
        # The signing slave's state can be 0 (full permissions) or 1 (mining only) but only 0 is used for now.
        # Let's give an invalid number.
        self.m_addr_from_pk_state.slave_pks_access_type = {str(tx.PK): 2}
        result = tx.validate_slave(self.m_addr_state, self.m_addr_from_pk_state)
        self.assertFalse(result)

        # Let's give a valid number, that matches what the master's state says (0)
        self.m_addr_from_pk_state.slave_pks_access_type = {str(tx.PK): 0}
        result = tx.validate_slave(self.m_addr_state, self.m_addr_from_pk_state)
        self.assertTrue(result)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from concurrent.futures import ThreadPoolExecutor

import grpc

from qrl.core import config
from qrl.core.misc import logger
from qrl.core.qrlnode import QRLNode
from qrl.generated.qrl_pb2_grpc import add_PublicAPIServicer_to_server, add_AdminAPIServicer_to_server
from qrl.generated.qrlmining_pb2_grpc import add_MiningAPIServicer_to_server
from qrl.generated.qrlbase_pb2_grpc import add_BaseServicer_to_server
from qrl.generated.qrldebug_pb2_grpc import add_DebugAPIServicer_to_server
from qrl.services.BaseService import BaseService
from qrl.services.AdminAPIService import AdminAPIService
from qrl.services.PublicAPIService import PublicAPIService
from qrl.services.MiningAPIService import MiningAPIService
from qrl.services.DebugAPIService import DebugAPIService


def start_services(node: QRLNode):
    public_server = grpc.server(ThreadPoolExecutor(max_workers=config.user.public_api_threads),
                                maximum_concurrent_rpcs=config.user.public_api_max_concurrent_rpc)
    add_BaseServicer_to_server(BaseService(node), public_server)
    add_PublicAPIServicer_to_server(PublicAPIService(node), public_server)

    if config.user.public_api_enabled:
        public_server.add_insecure_port("{0}:{1}".format(config.user.public_api_host,
                                                         config.user.public_api_port))
        public_server.start()

        logger.info("grpc public service - started !")

    admin_server = grpc.server(ThreadPoolExecutor(max_workers=config.user.admin_api_threads),
                               maximum_concurrent_rpcs=config.user.admin_api_max_concurrent_rpc)
    add_AdminAPIServicer_to_server(AdminAPIService(node), admin_server)

    if config.user.admin_api_enabled:
        admin_server.add_insecure_port("{0}:{1}".format(config.user.admin_api_host,
                                                        config.user.admin_api_port))
        admin_server.start()

        logger.info("grpc admin service - started !")

    mining_server = grpc.server(ThreadPoolExecutor(max_workers=config.user.mining_api_threads),
                                maximum_concurrent_rpcs=config.user.mining_api_max_concurrent_rpc)
    add_MiningAPIServicer_to_server(MiningAPIService(node), mining_server)

    if config.user.mining_api_enabled:
        mining_server.add_insecure_port("{0}:{1}".format(config.user.mining_api_host,
                                                         config.user.mining_api_port))
        mining_server.start()

        logger.info("grpc mining service - started !")

    debug_server = grpc.server(ThreadPoolExecutor(max_workers=config.user.debug_api_threads),
                               maximum_concurrent_rpcs=config.user.debug_api_max_concurrent_rpc)
    add_DebugAPIServicer_to_server(DebugAPIService(node), debug_server)

    if config.user.debug_api_enabled:
        debug_server.add_insecure_port("{0}:{1}".format(config.user.debug_api_host,
                                                        config.user.debug_api_port))
        debug_server.start()

        logger.info("grpc debug service - started !")

    return admin_server, public_server, mining_server, debug_server



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from pyqrllib.pyqrllib import bin2hstr
from qrl.generated import qrlwallet_pb2
from qrl.generated.qrlwallet_pb2_grpc import WalletAPIServicer
from qrl.services.grpcHelper import GrpcExceptionWrapper


class WalletAPIService(WalletAPIServicer):
    MAX_REQUEST_QUANTITY = 100

    # TODO: Separate the Service from the node model
    def __init__(self, walletd):
        self._walletd = walletd

    @GrpcExceptionWrapper(qrlwallet_pb2.AddNewAddressResp)
    def AddNewAddress(self, request: qrlwallet_pb2.AddNewAddressReq, context) -> qrlwallet_pb2.AddNewAddressResp:
        resp = qrlwallet_pb2.AddNewAddressResp()
        try:
            resp.address = self._walletd.add_new_address(request.height, request.hash_function.lower())
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.AddNewAddressResp)
    def AddNewAddressWithSlaves(self, request: qrlwallet_pb2.AddNewAddressWithSlavesReq, context) -> qrlwallet_pb2.AddNewAddressResp:
        resp = qrlwallet_pb2.AddNewAddressResp()
        try:
            resp.address = self._walletd.add_new_address_with_slaves(request.height,
                                                                     request.number_of_slaves,
                                                                     request.hash_function.lower())
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.AddAddressFromSeedResp)
    def AddAddressFromSeed(self, request: qrlwallet_pb2.AddAddressFromSeedReq, context) -> qrlwallet_pb2.AddAddressFromSeedResp:
        resp = qrlwallet_pb2.AddAddressFromSeedResp()
        try:
            resp.address = self._walletd.add_address_from_seed(seed=request.seed)
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.ListAddressesResp)
    def ListAddresses(self, request: qrlwallet_pb2.ListAddressesReq, context) -> qrlwallet_pb2.ListAddressesResp:
        resp = qrlwallet_pb2.ListAddressesResp()
        try:
            resp.addresses.extend(self._walletd.list_address())
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.RemoveAddressResp)
    def RemoveAddress(self, request: qrlwallet_pb2.RemoveAddressReq, context) -> qrlwallet_pb2.RemoveAddressResp:
        resp = qrlwallet_pb2.RemoveAddressResp()
        try:
            if not self._walletd.remove_address(request.address):
                resp.code = 1
                resp.error = "No such address found"
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.ValidAddressResp)
    def IsValidAddress(self, request: qrlwallet_pb2.ValidAddressReq, context) -> qrlwallet_pb2.ValidAddressResp:
        resp = qrlwallet_pb2.ValidAddressResp()
        try:
            if not self._walletd.validate_address(request.address):
                resp.code = 1
                resp.error = "Invalid QRL Address"
                resp.valid = "False"
            else:
                resp.valid = "True"
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.GetRecoverySeedsResp)
    def GetRecoverySeeds(self, request: qrlwallet_pb2.GetRecoverySeedsReq, context) -> qrlwallet_pb2.GetRecoverySeedsResp:
        resp = qrlwallet_pb2.GetRecoverySeedsResp()
        try:
            resp.hexseed, resp.mnemonic = self._walletd.get_recovery_seeds(request.address)
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.GetWalletInfoResp)
    def GetWalletInfo(self, request: qrlwallet_pb2.GetWalletInfoReq, context) -> qrlwallet_pb2.GetWalletInfoResp:
        resp = qrlwallet_pb2.GetWalletInfoResp()
        try:
            resp.version, resp.address_count, resp.is_encrypted = self._walletd.get_wallet_info()
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.RelayTxnResp)
    def RelayTransferTxn(self, request: qrlwallet_pb2.RelayTransferTxnReq, context) -> qrlwallet_pb2.RelayTxnResp:
        resp = qrlwallet_pb2.RelayTxnResp()
        try:
            resp.tx.MergeFrom(self._walletd.relay_transfer_txn(request.addresses_to,
                                                               request.amounts,
                                                               request.fee,
                                                               request.master_address,
                                                               request.signer_address,
                                                               request.ots_index))
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.RelayTxnResp)
    def RelayTransferTxnBySlave(self,
                                request: qrlwallet_pb2.RelayTransferTxnBySlaveReq,
                                context) -> qrlwallet_pb2.RelayTxnResp:
        resp = qrlwallet_pb2.RelayTxnResp()
        try:
            resp.tx.MergeFrom(self._walletd.relay_transfer_txn_by_slave(request.addresses_to,
                                                                        request.amounts,
                                                                        request.fee,
                                                                        request.master_address))
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.RelayTxnResp)
    def RelayMessageTxn(self, request: qrlwallet_pb2.RelayMessageTxnReq, context) -> qrlwallet_pb2.RelayTxnResp:
        resp = qrlwallet_pb2.RelayTxnResp()
        try:
            resp.tx.MergeFrom(self._walletd.relay_message_txn(request.message,
                                                              request.fee,
                                                              request.master_address,
                                                              request.signer_address,
                                                              request.ots_index))
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.RelayTxnResp)
    def RelayMessageTxnBySlave(self,
                               request: qrlwallet_pb2.RelayMessageTxnBySlaveReq,
                               context) -> qrlwallet_pb2.RelayTxnResp:
        resp = qrlwallet_pb2.RelayTxnResp()
        try:
            resp.tx.MergeFrom(self._walletd.relay_message_txn_by_slave(request.message,
                                                                       request.fee,
                                                                       request.master_address))
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.RelayTxnResp)
    def RelayTokenTxn(self, request: qrlwallet_pb2.RelayTokenTxnReq, context) -> qrlwallet_pb2.RelayTxnResp:
        resp = qrlwallet_pb2.RelayTxnResp()
        try:
            resp.tx.MergeFrom(self._walletd.relay_token_txn(request.symbol,
                                                            request.name,
                                                            request.owner,
                                                            request.decimals,
                                                            request.addresses,
                                                            request.amounts,
                                                            request.fee,
                                                            request.master_address,
                                                            request.signer_address,
                                                            request.ots_index))
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.RelayTxnResp)
    def RelayTokenTxnBySlave(self,
                             request: qrlwallet_pb2.RelayTokenTxnBySlaveReq,
                             context) -> qrlwallet_pb2.RelayTxnResp:
        resp = qrlwallet_pb2.RelayTxnResp()
        try:
            resp.tx.MergeFrom(self._walletd.relay_token_txn_by_slave(request.symbol,
                                                                     request.name,
                                                                     request.owner,
                                                                     request.decimals,
                                                                     request.addresses,
                                                                     request.amounts,
                                                                     request.fee,
                                                                     request.master_address))
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.RelayTxnResp)
    def RelayTransferTokenTxn(self, request: qrlwallet_pb2.RelayTransferTokenTxnReq, context) -> qrlwallet_pb2.RelayTxnResp:
        resp = qrlwallet_pb2.RelayTxnResp()
        try:
            resp.tx.MergeFrom(self._walletd.relay_transfer_token_txn(request.addresses_to,
                                                                     request.amounts,
                                                                     request.token_txhash,
                                                                     request.fee,
                                                                     request.master_address,
                                                                     request.signer_address,
                                                                     request.ots_index))
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.RelayTxnResp)
    def RelayTransferTokenTxnBySlave(self,
                                     request: qrlwallet_pb2.RelayTransferTokenTxnBySlaveReq,
                                     context) -> qrlwallet_pb2.RelayTxnResp:
        resp = qrlwallet_pb2.RelayTxnResp()
        try:
            resp.tx.MergeFrom(self._walletd.relay_transfer_token_txn_by_slave(request.addresses_to,
                                                                              request.amounts,
                                                                              request.token_txhash,
                                                                              request.fee,
                                                                              request.master_address))
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.RelayTxnResp)
    def RelaySlaveTxn(self, request: qrlwallet_pb2.RelaySlaveTxnReq, context) -> qrlwallet_pb2.RelayTxnResp:
        resp = qrlwallet_pb2.RelayTxnResp()
        try:
            resp.tx.MergeFrom(self._walletd.relay_slave_txn(request.slave_pks,
                                                            request.access_types,
                                                            request.fee,
                                                            request.master_address,
                                                            request.signer_address,
                                                            request.ots_index))
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.RelayTxnResp)
    def RelaySlaveTxnBySlave(self, request: qrlwallet_pb2.RelaySlaveTxnBySlaveReq, context) -> qrlwallet_pb2.RelayTxnResp:
        resp = qrlwallet_pb2.RelayTxnResp()
        try:
            resp.tx.MergeFrom(self._walletd.relay_slave_txn_by_slave(request.slave_pks,
                                                                     request.access_types,
                                                                     request.fee,
                                                                     request.master_address))
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.EncryptWalletResp)
    def EncryptWallet(self, request: qrlwallet_pb2.EncryptWalletReq, context) -> qrlwallet_pb2.EncryptWalletResp:
        resp = qrlwallet_pb2.EncryptWalletResp()
        try:
            self._walletd.encrypt_wallet(request.passphrase)
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.LockWalletResp)
    def LockWallet(self, request: qrlwallet_pb2.LockWalletReq, context) -> qrlwallet_pb2.LockWalletResp:
        resp = qrlwallet_pb2.LockWalletResp()
        try:
            self._walletd.lock_wallet()
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.UnlockWalletResp)
    def UnlockWallet(self, request: qrlwallet_pb2.UnlockWalletReq, context) -> qrlwallet_pb2.UnlockWalletResp:
        resp = qrlwallet_pb2.UnlockWalletResp()
        try:
            self._walletd.unlock_wallet(request.passphrase)
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.ChangePassphraseResp)
    def ChangePassphrase(self,
                         request: qrlwallet_pb2.ChangePassphraseReq,
                         context) -> qrlwallet_pb2.ChangePassphraseResp:
        resp = qrlwallet_pb2.ChangePassphraseResp()
        try:
            self._walletd.change_passphrase(request.oldPassphrase, request.newPassphrase)
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.TransactionsByAddressResp)
    def GetTransactionsByAddress(self,
                                 request: qrlwallet_pb2.TransactionsByAddressReq,
                                 context) -> qrlwallet_pb2.TransactionsByAddressResp:
        resp = qrlwallet_pb2.TransactionsByAddressResp()
        try:
            mini_transactions, balance = self._walletd.get_transactions_by_address(request.address)
            resp.mini_transactions.extend(mini_transactions)
            resp.balance = balance
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.TransactionResp)
    def GetTransaction(self, request: qrlwallet_pb2.TransactionReq, context) -> qrlwallet_pb2.TransactionResp:
        resp = qrlwallet_pb2.TransactionResp()
        try:
            tx, confirmations, block_number, block_header_hash = self._walletd.get_transaction(request.tx_hash)
            resp.tx.MergeFrom(tx)
            resp.confirmations = confirmations
            resp.block_number = block_number
            if block_header_hash:
                resp.block_header_hash = block_header_hash
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.BalanceResp)
    def GetBalance(self, request: qrlwallet_pb2.BalanceReq, context) -> qrlwallet_pb2.BalanceResp:
        resp = qrlwallet_pb2.BalanceResp()
        try:
            resp.balance = str(self._walletd.get_balance(request.address))
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.OTSResp)
    def GetOTS(self, request: qrlwallet_pb2.OTSReq, context) -> qrlwallet_pb2.OTSResp:
        resp = qrlwallet_pb2.OTSResp()
        try:
            ots_bitfield, next_unused_ots_index = self._walletd.get_ots(request.address)
            resp.ots_bitfield.extend(ots_bitfield)
            resp.next_unused_ots_index = next_unused_ots_index
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.HeightResp)
    def GetHeight(self, request: qrlwallet_pb2.HeightReq, context) -> qrlwallet_pb2.HeightResp:
        resp = qrlwallet_pb2.HeightResp()
        try:
            resp.height = self._walletd.get_height()
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.BlockResp)
    def GetBlock(self, request: qrlwallet_pb2.BlockReq, context) -> qrlwallet_pb2.BlockResp:
        resp = qrlwallet_pb2.BlockResp()
        try:
            resp.block.MergeFrom(self._walletd.get_block(request.header_hash))
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.BlockResp)
    def GetBlockByNumber(self, request: qrlwallet_pb2.BlockByNumberReq, context) -> qrlwallet_pb2.BlockResp:
        resp = qrlwallet_pb2.BlockResp()
        try:
            resp.block.MergeFrom(self._walletd.get_block_by_number(request.block_number))
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.AddressFromPKResp)
    def GetAddressFromPK(self, request: qrlwallet_pb2.AddressFromPKReq, context) -> qrlwallet_pb2.AddressFromPKResp:
        resp = qrlwallet_pb2.AddressFromPKResp()
        try:
            resp.address = self._walletd.get_address_from_pk(request.pk)
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp

    @GrpcExceptionWrapper(qrlwallet_pb2.NodeInfoResp)
    def GetNodeInfo(self, request: qrlwallet_pb2.NodeInfoReq, context) -> qrlwallet_pb2.NodeInfoResp:
        resp = qrlwallet_pb2.NodeInfoResp()
        try:
            node_info = self._walletd.get_node_info()

            resp.version = node_info.info.version
            resp.num_connections = str(node_info.info.num_connections)
            resp.num_known_peers = str(node_info.info.num_known_peers)
            resp.uptime = node_info.info.uptime
            resp.block_height = node_info.info.block_height
            resp.block_last_hash = bin2hstr(node_info.info.block_last_hash)
            resp.network_id = node_info.info.network_id
        except Exception as e:
            resp.code = 1
            resp.error = str(e)

        return resp



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from random import shuffle
from unittest import TestCase
from mock import Mock, PropertyMock, patch

from tests.misc.helper import get_alice_xmss, get_slave_xmss, get_random_xmss
from qrl.core.misc import logger
from qrl.core import config
from qrl.core.AddressState import AddressState

logger.initialize_default()

alice = get_alice_xmss()
slave = get_slave_xmss()


class TestAddressState(TestCase):
    def setUp(self):
        self.addr_state = AddressState.get_default(alice.address)

    def test_create_and_properties(self):
        a = AddressState.create(address=alice.address, nonce=0, balance=10,
                                ots_bitfield=[b'\x00'] * config.dev.ots_bitfield_size,
                                tokens={b'010101': 100, b'020202': 200},
                                slave_pks_access_type={slave.pk: 1},
                                ots_counter=0
                                )
        self.assertEqual(a.pbdata.address, a.address)
        self.assertEqual(a.pbdata.balance, a.balance)
        a.balance = 3
        self.assertEqual(a.balance, 3)
        self.assertEqual(a.pbdata.nonce, a.nonce)
        self.assertEqual(a.pbdata.ots_bitfield, a.ots_bitfield)
        self.assertEqual(a.pbdata.ots_counter, a.ots_counter)
        self.assertEqual(a.pbdata.transaction_hashes, a.transaction_hashes)
        self.assertEqual(a.pbdata.latticePK_list, a.latticePK_list)
        self.assertEqual(a.pbdata.slave_pks_access_type, a.slave_pks_access_type)

    def test_token_balance_functionality(self):
        # If I update an AddressState's token balance, it should do what the function name says.
        self.addr_state.update_token_balance(b'010101', 10)
        self.assertEqual(self.addr_state.get_token_balance(b'010101'), 10)
        self.assertTrue(self.addr_state.is_token_exists(b'010101'))

        # I can call update_token_balance with a negative number to decrease the balance.
        self.addr_state.update_token_balance(b'010101', -2)
        self.assertEqual(self.addr_state.get_token_balance(b'010101'), 8)

        # If the token balance hits 0, the token_txhash should have been pruned from the AddressState.
        # And when I ask for its balance, it should return 0.
        self.addr_state.update_token_balance(b'010101', -8)
        self.assertFalse(self.addr_state.is_token_exists(b'010101'))
        self.assertEqual(self.addr_state.get_token_balance(b'010101'), 0)

    def test_nonce(self):
        self.addr_state.increase_nonce()
        self.assertEqual(self.addr_state.nonce, 1)

        self.addr_state.increase_nonce()
        self.addr_state.increase_nonce()
        self.assertEqual(self.addr_state.nonce, 3)

        self.addr_state.decrease_nonce()
        self.addr_state.decrease_nonce()
        self.assertEqual(self.addr_state.nonce, 1)

    def test_nonce_negative(self):
        with self.assertRaises(ValueError):
            self.addr_state.decrease_nonce()

    def test_slave_pks_access_type(self):
        # slave_pks_access_type could take 2 values: 0 (all permission granted to slaves), 1 (only mining)
        # There is no validation for the values of slave_pks_access_type.
        # For now only 0 is used.
        # By default all slave_pks get permission level 0
        self.addr_state.add_slave_pks_access_type(slave.pk, 1)
        self.assertEqual(self.addr_state.slave_pks_access_type[str(slave.pk)], 1)

    def test_get_slave_permission(self):
        # We haven't added slave.pk to the addr_state yet, so slave is not yet a slave of this AddressState.
        self.assertEqual(self.addr_state.get_slave_permission(slave.pk), -1)

        # Add slave permissions for slave.pk
        self.addr_state.add_slave_pks_access_type(slave.pk, 1)
        self.assertEqual(self.addr_state.get_slave_permission(slave.pk), 1)

        # Remove slave permissions for slave.pk
        self.addr_state.remove_slave_pks_access_type(slave.pk)
        self.assertEqual(self.addr_state.get_slave_permission(slave.pk), -1)

    def test_lattice_txn(self):
        # Add a mock LatticeTransaction to the AddressState. It should show up in addr_state.latticePK_list
        m_lattice_txn = Mock(name='mock LatticeTransaction', txhash=b'txhash', dilithium_pk=b'dilithiumPK',
                             kyber_pk=b'kyberPK')

        self.assertEqual(len(self.addr_state.latticePK_list), 0)
        self.addr_state.add_lattice_pk(m_lattice_txn)
        self.assertEqual(len(self.addr_state.latticePK_list), 1)

        # Add another different LatticeTransaction. addr_state.latticePK_list should now be longer.
        m_lattice_txn_2 = Mock(name='mock LatticeTransaction 2', txhash=b'deadbeef', dilithium_pk=b'dilithiumPK',
                               kyber_pk=b'kyberPK')
        self.addr_state.add_lattice_pk(m_lattice_txn_2)
        self.assertEqual(len(self.addr_state.latticePK_list), 2)

        # Remove the second LatticeTransaction.
        self.addr_state.remove_lattice_pk(m_lattice_txn_2)
        self.assertEqual(len(self.addr_state.latticePK_list), 1)

        # Try to remove the second LatticeTransaction again
        # The list should remain the same.
        self.addr_state.remove_lattice_pk(m_lattice_txn_2)
        self.assertEqual(len(self.addr_state.latticePK_list), 1)

    def test_get_default_coinbase(self):
        # Make sure that Coinbase AddressState gets all the coins supply by default
        coinbase_addr_state = AddressState.get_default(config.dev.coinbase_address)
        self.assertEqual(coinbase_addr_state.balance, int(config.dev.max_coin_supply * config.dev.shor_per_quanta))

    def test_set_ots_key(self):
        # If it's below config.dev.max_ots_tracking_index, use the bitfield.
        self.assertEqual(b'\x00', self.addr_state.ots_bitfield[0])
        self.addr_state.set_ots_key(0)
        self.assertEqual(b'\x01', self.addr_state.ots_bitfield[0])
        self.assertEqual(self.addr_state.ots_counter, 0)

        self.assertEqual(b'\x00', self.addr_state.ots_bitfield[-1])
        self.addr_state.set_ots_key(config.dev.max_ots_tracking_index - 1)
        self.assertEqual(b'\x80', self.addr_state.ots_bitfield[-1])

        # Start using the counter from config.dev.max_ots_tracking_index and above.
        self.addr_state.set_ots_key(config.dev.max_ots_tracking_index)
        self.assertEqual(self.addr_state.ots_counter, config.dev.max_ots_tracking_index)
        self.addr_state.set_ots_key(config.dev.max_ots_tracking_index + 1)
        self.assertEqual(self.addr_state.ots_counter, config.dev.max_ots_tracking_index + 1)

    def test_unset_ots_key(self):
        m_state = Mock(name='mock State')
        # Set the OTS bitfield, and then try to unset it.
        self.addr_state.set_ots_key(0)
        self.addr_state.unset_ots_key(0, m_state)
        self.assertEqual(self.addr_state.ots_bitfield[0], b'\x00')

    @patch('qrl.core.AddressState.AddressState.transaction_hashes', new_callable=PropertyMock)
    def test_unset_ots_key_counter(self, m_transaction_hashes):
        # unset_ots_key() walks backwards through transaction_hashes and finds the last txhash that used OTS counter
        # index.
        # This assumes that transaction_hashes is sorted in time-ascending order, and that the offending tx
        # has already been removed from transaction_hashes.
        m_transaction_hashes.return_value = ["1st hash", "2nd hash", "3rd hash", "4th hash"]
        m_state = Mock(name='mock State')

        # The user has used some OTS bitfield indexes in a random fashion, and has two OTS counter transactions.
        m_state.get_tx_metadata.side_effect = [
            (Mock(name="4th hash", ots_key=357), "unused"),
            (Mock(name="3rd hash", ots_key=config.dev.max_ots_tracking_index + 3), "unused"),
            (Mock(name="2nd hash", ots_key=952), "unused"),
            (Mock(name="1st hash", ots_key=config.dev.max_ots_tracking_index + 2), "unused")
        ]
        self.addr_state.set_ots_key(config.dev.max_ots_tracking_index + 4)  # 8196

        # You must remove the txhash that uses index 8196 from transaction_hashes BEFORE calling this function, as it
        # will walk backwards through the transaction_hashes
        self.addr_state.unset_ots_key(config.dev.max_ots_tracking_index + 4, m_state)

        self.assertEqual(self.addr_state.ots_counter, config.dev.max_ots_tracking_index + 3)

    @patch('qrl.core.AddressState.AddressState.transaction_hashes', new_callable=PropertyMock)
    def test_unset_ots_key_counter_unsorted_transaction_hashes(self, m_transaction_hashes):
        # unset_ots_key() ends up with the wrong OTS counter index if transaction_hashes is not properly sorted.
        m_transaction_hashes.return_value = ["1st hash", "2nd hash", "3rd hash", "4th hash"]
        m_state = Mock(name='mock State')

        # The user has used some OTS bitfield indexes in a random fashion, and has two OTS counter transactions.
        m_state.get_tx_metadata.side_effect = [
            (Mock(name="4th hash", ots_key=357), "unused"),
            (Mock(name="3rd hash", ots_key=config.dev.max_ots_tracking_index + 2), "unused"),
            (Mock(name="2nd hash", ots_key=952), "unused"),
            (Mock(name="1st hash", ots_key=config.dev.max_ots_tracking_index + 3), "unused")
        ][-1::-1]  # reversing the order, as it will be called in reverse order by unset_ots_key
        self.addr_state.set_ots_key(config.dev.max_ots_tracking_index + 4)  # 8196

        self.addr_state.unset_ots_key(config.dev.max_ots_tracking_index + 4, m_state)

        # In a perfect world, unset_ots_key should have found "1st hash" and used its ots_key
        # But it found "3rd hash" and used its ots key, thus FAIL.
        self.assertEqual(self.addr_state.ots_counter, config.dev.max_ots_tracking_index + 3)

    def test_ots_key_validation(self):
        random_xmss = get_random_xmss(xmss_height=4)
        addr = AddressState.get_default(random_xmss.address)
        ots_indexes = list(range(0, 2 ** random_xmss.height))
        shuffle(ots_indexes)

        for i in ots_indexes:
            if i < config.dev.max_ots_tracking_index:
                self.assertFalse(addr.ots_key_reuse(i))
            else:
                result = addr.ots_key_reuse(i)
                if i > addr.ots_counter:
                    self.assertFalse(result)
                else:
                    self.assertTrue(result)

            addr.set_ots_key(i)

            self.assertTrue(addr.ots_key_reuse(i))

    def test_ots_key_reuse_counter(self):
        # If the ots counter is 8193, that automatically means index 8192 is used.
        self.addr_state.set_ots_key(config.dev.max_ots_tracking_index + 1)
        self.assertTrue(self.addr_state.ots_key_reuse(config.dev.max_ots_tracking_index))
        # If the ots counter is 8193, that means 8194 and above are not used.
        self.assertFalse(self.addr_state.ots_key_reuse(config.dev.max_ots_tracking_index + 2))

    def test_serialize(self):
        # Simply test that serialize() works and you can deserialize from it.
        output = self.addr_state.serialize()
        another_addr_state = AddressState(protobuf_block=output)
        self.assertIsInstance(another_addr_state, AddressState)

    def test_address_is_valid(self):
        self.assertTrue(AddressState.address_is_valid(alice.address))
        self.assertFalse(AddressState.address_is_valid(b'fake address'))

    def test_height(self):
        self.assertEqual(self.addr_state.height, alice.address[1] << 1)

    def test_get_unused_ots_index(self):
        random_xmss = get_random_xmss(xmss_height=4)
        addr_state = AddressState.get_default(random_xmss.address)
        self.assertEqual(addr_state.get_unused_ots_index(), 0)
        addr_state.set_ots_key(0)
        self.assertEqual(addr_state.get_unused_ots_index(), 1)
        addr_state.set_ots_key(2)
        self.assertEqual(addr_state.get_unused_ots_index(), 1)
        addr_state.set_ots_key(1)
        self.assertEqual(addr_state.get_unused_ots_index(), 3)
        for i in range(3, min(2 ** addr_state.height, config.dev.max_ots_tracking_index)):
            addr_state.set_ots_key(i)
        self.assertIsNone(addr_state.get_unused_ots_index())

    def test_get_unused_ots_index2(self):
        old_value = config.dev.max_ots_tracking_index
        config.dev.max_ots_tracking_index = 128

        try:
            random_xmss = get_random_xmss(xmss_height=8)
            addr_state = AddressState.get_default(random_xmss.address)
            self.assertEqual(addr_state.get_unused_ots_index(), 0)
            addr_state.set_ots_key(0)
            self.assertEqual(addr_state.get_unused_ots_index(), 1)
            addr_state.set_ots_key(2)
            self.assertEqual(addr_state.get_unused_ots_index(), 1)
            addr_state.set_ots_key(1)
            self.assertEqual(addr_state.get_unused_ots_index(), 3)

            for i in range(3, min(2 ** addr_state.height, config.dev.max_ots_tracking_index)):
                addr_state.set_ots_key(i)
                self.assertEqual(addr_state.get_unused_ots_index(), i + 1)

            self.assertEqual(addr_state.get_unused_ots_index(), config.dev.max_ots_tracking_index)

            for i in range(config.dev.max_ots_tracking_index, 2 ** addr_state.height):
                addr_state.set_ots_key(i)

            self.assertIsNone(addr_state.get_unused_ots_index())
        finally:
            config.dev.max_ots_tracking_index = old_value



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import threading
from unittest import TestCase

from pyqryptonight.pyqryptonight import Qryptominer, PoWHelper, SOLUTION

from qrl.core import config
from qrl.core.misc import logger
from qrl.core.Block import Block
from qrl.core.DifficultyTracker import DifficultyTracker
from tests.misc.helper import read_data_file

logger.initialize_default()


class TestQryptominer(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestQryptominer, self).__init__(*args, **kwargs)

    def test_verify(self):

        class CustomQMiner(Qryptominer):
            def __init__(self):
                Qryptominer.__init__(self)
                self._solution_lock = threading.Lock()
                self.nonce = None
                self.solution_blob = None

            def start(self, input, nonceOffset, target, thread_count):
                self.cancel()
                try:
                    self._solution_lock.release()
                except RuntimeError:
                    pass
                self._solution_lock.acquire(blocking=False)
                super().start(input, nonceOffset, target, thread_count)

            def wait_for_solution(self):
                self._solution_lock.acquire(blocking=True)
                self._solution_lock.release()

            def handleEvent(self, event):
                if event.type == SOLUTION:
                    self.nonce = event.nonce
                    self.solution_blob = self.solutionInput()
                    self._solution_lock.release()

        block_timestamp = 1515443508
        parent_block_timestamp = 1515443508

        # This could be the average of last N blocks
        measurement = block_timestamp - parent_block_timestamp

        parent_difficulty = (0, 0, 0, 0, 0, 0, 0, 0,
                             0, 0, 0, 0, 0, 0, 0, 0,
                             0, 0, 0, 0, 0, 0, 0, 0,
                             0, 0, 0, 0, 0, 0, 0, 4)

        new_diff, new_target = DifficultyTracker.get(
            measurement,
            parent_difficulty=parent_difficulty)

        self.assertEqual(new_diff, (0, 0, 0, 0, 0, 0, 0, 0,
                                    0, 0, 0, 0, 0, 0, 0, 0,
                                    0, 0, 0, 0, 0, 0, 0, 0,
                                    0, 0, 0, 0, 0, 0, 0, 5))

        self.assertEqual(new_target, (
            51, 51, 51, 51, 51, 51, 51, 51,
            51, 51, 51, 51, 51, 51, 51, 51,
            51, 51, 51, 51, 51, 51, 51, 51,
            51, 51, 51, 51, 51, 51, 51, 51))

        block_json = read_data_file('core/example_block_mining.json')

        block = Block.from_json(block_json)

        expected_blob = (0, 231, 90, 101, 142, 20, 245, 183, 96, 5, 216, 159, 111, 239, 93, 217, 138, 10, 227, 159, 198,
                         207, 109, 238, 83, 220, 167, 148, 247, 200, 197, 41, 37, 36, 150, 12, 116, 85, 254, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 181, 198, 40, 62, 106, 139, 108, 83, 216, 206, 161, 148,
                         50, 65, 212, 137, 94, 102, 124, 45)
        self.assertEqual(expected_blob, tuple(block.mining_blob))

        custom_qminer = CustomQMiner()
        custom_qminer.start(input=block.mining_blob,
                            nonceOffset=block.mining_nonce_offset,
                            target=new_target,
                            thread_count=2)
        custom_qminer.wait_for_solution()

        expected_mined_blob = bytearray(expected_blob)
        tmp_offset = config.dev.mining_nonce_offset
        expected_mined_blob[tmp_offset:tmp_offset + 4] = custom_qminer.nonce.to_bytes(4,
                                                                                      byteorder='big',
                                                                                      signed=False)

        print(custom_qminer.nonce)
        self.assertEqual(tuple(expected_mined_blob), custom_qminer.solution_blob)
        self.assertTrue(PoWHelper().verifyInput(custom_qminer.solution_blob, new_target))



from unittest import TestCase

import simplejson as json
from mock import patch, PropertyMock, Mock
from pyqrllib.pyqrllib import bin2hstr

from qrl.core.misc import logger
from qrl.core.AddressState import AddressState
from qrl.core.txs.TokenTransaction import TokenTransaction
from qrl.core.txs.Transaction import Transaction
from qrl.generated import qrl_pb2
from tests.core.txs.testdata import test_json_Token, test_signature_Token
from tests.misc.helper import get_alice_xmss, get_bob_xmss, get_slave_xmss

logger.initialize_default()


@patch('qrl.core.txs.Transaction.logger')
class TestTokenTransaction(TestCase):

    def __init__(self, *args, **kwargs):
        super(TestTokenTransaction, self).__init__(*args, **kwargs)
        self.alice = get_alice_xmss()
        self.bob = get_bob_xmss()
        self._decimals = 15

        self.alice.set_ots_index(10)
        self.maxDiff = None

    def setUp(self):
        self.initial_balances_valid = [qrl_pb2.AddressAmount(address=self.alice.address, amount=1000),
                                       qrl_pb2.AddressAmount(address=self.bob.address, amount=1000)]

        self.params = {"symbol": b'QRL',
                       "name": b'Quantum Resistant Ledger',
                       "owner": self.alice.address,
                       "decimals": self._decimals,
                       "initial_balances": self.initial_balances_valid,
                       "fee": 1,
                       "xmss_pk": self.alice.pk}

    def make_tx(self, **kwargs):
        self.params.update(kwargs)
        tx = TokenTransaction.create(**self.params)
        return tx

    def test_create(self, m_logger):
        # Alice creates Token
        initial_balances = list()
        initial_balances.append(qrl_pb2.AddressAmount(address=self.alice.address,
                                                      amount=400000000))
        initial_balances.append(qrl_pb2.AddressAmount(address=self.bob.address,
                                                      amount=200000000))

        tx = self.make_tx(decimals=4, initial_balances=initial_balances)

        self.assertTrue(tx)

    def test_create_negative_fee(self, m_logger):
        with self.assertRaises(ValueError):
            TokenTransaction.create(symbol=b'QRL',
                                    name=b'Quantum Resistant Ledger',
                                    owner=self.alice.address,
                                    decimals=4,
                                    initial_balances=[],
                                    fee=-1,
                                    xmss_pk=self.alice.pk)

    def test_to_json(self, m_logger):
        initial_balances = list()
        initial_balances.append(qrl_pb2.AddressAmount(address=self.alice.address,
                                                      amount=400000000))
        initial_balances.append(qrl_pb2.AddressAmount(address=self.bob.address,
                                                      amount=200000000))
        tx = TokenTransaction.create(symbol=b'QRL',
                                     name=b'Quantum Resistant Ledger',
                                     owner=b'\x01\x03\x17F=\xcdX\x1bg\x9bGT\xf4ld%\x12T\x89\xa2\x82h\x94\xe3\xc4*Y\x0e\xfbh\x06E\x0c\xe6\xbfRql',
                                     decimals=4,
                                     initial_balances=initial_balances,
                                     fee=1,
                                     xmss_pk=self.alice.pk)
        txjson = tx.to_json()

        self.assertEqual(json.loads(test_json_Token), json.loads(txjson))

    def test_from_json(self, m_logger):
        tx = Transaction.from_json(test_json_Token)
        tx.sign(self.alice)
        self.assertIsInstance(tx, TokenTransaction)

        # Test that common Transaction components were copied over.
        self.assertEqual('010300a1da274e68c88b0ccf448e0b1916fa789b01eb2ed4e9ad565ce264c9390782a9c61ac02f',
                         bin2hstr(tx.addr_from))
        self.assertEqual('01030038ea6375069f8272cc1a6601b3c76c21519455603d370036b97c779ada356'
                         '5854e3983bd564298c49ae2e7fa6e28d4b954d8cd59398f1225b08d6144854aee0e',
                         bin2hstr(tx.PK))
        self.assertEqual(b'QRL', tx.symbol)
        self.assertEqual(b'Quantum Resistant Ledger', tx.name)
        self.assertEqual('010317463dcd581b679b4754f46c6425125489a2826894e3c42a590efb6806450ce6bf52716c',
                         bin2hstr(tx.owner))
        self.assertEqual('ff84da605e9c9cd04d68503be7922110b4cc147837f8687ad18aa54b7bc5632d', bin2hstr(tx.txhash))
        self.assertEqual(10, tx.ots_key)

        self.assertEqual(test_signature_Token, bin2hstr(tx.signature))

        total_supply = 0
        for initial_balance in tx.initial_balances:
            total_supply += initial_balance.amount
        self.assertEqual(600000000, total_supply)

        self.assertEqual(1, tx.fee)

    def test_validate_tx(self, m_logger):
        initial_balances = list()
        initial_balances.append(qrl_pb2.AddressAmount(address=self.alice.address,
                                                      amount=400000000))
        initial_balances.append(qrl_pb2.AddressAmount(address=self.bob.address,
                                                      amount=200000000))

        tx = self.make_tx(decimals=4, initial_balances=initial_balances)

        # We must sign the tx before validation will work.
        tx.sign(self.alice)

        # We have not touched the tx: validation should pass.
        self.assertTrue(tx.validate_or_raise())

    def test_validate_tx2(self, m_logger):
        initial_balances = list()
        initial_balances.append(qrl_pb2.AddressAmount(address=self.alice.address,
                                                      amount=10000000000000000001))
        initial_balances.append(qrl_pb2.AddressAmount(address=self.bob.address,
                                                      amount=10000000000000000001))

        # Transaction Validation should fail as the decimals is higher than the possible decimals
        with self.assertRaises(ValueError):
            self.make_tx(decimals=4, initial_balances=initial_balances)

    def test_validate_tx3(self, m_logger):
        initial_balances = list()
        initial_balances.append(qrl_pb2.AddressAmount(address=self.alice.address,
                                                      amount=1000 * 10 ** self._decimals))
        initial_balances.append(qrl_pb2.AddressAmount(address=self.bob.address,
                                                      amount=1000 * 10 ** self._decimals))

        tx = self.make_tx(initial_balances=initial_balances)

        # We must sign the tx before validation will work.
        tx.sign(self.alice)

        # We have not touched the tx: validation should pass.
        self.assertTrue(tx.validate_or_raise())

    def test_validate_tx4(self, m_logger):
        initial_balances = list()
        initial_balances.append(qrl_pb2.AddressAmount(address=self.alice.address,
                                                      amount=1000 * 10 ** self._decimals))
        initial_balances.append(qrl_pb2.AddressAmount(address=self.bob.address,
                                                      amount=1000 * 10 ** self._decimals))

        tx = self.make_tx(initial_balances=initial_balances)

        tx.sign(self.alice)

        self.assertTrue(tx.validate_or_raise())

        tx._data.transaction_hash = b'abc'

        with self.assertRaises(ValueError):
            tx.validate_or_raise()

    def test_validate_custom(self, m_logger):
        # Token symbol too long
        with self.assertRaises(ValueError):
            tx = self.make_tx(symbol=b'QRLSQRLSQRL')
            tx.sign(self.alice)

        # Token name too long
        with self.assertRaises(ValueError):
            tx = self.make_tx(name=b'Quantum Resistant LedgerQuantum')
            tx.sign(self.alice)

        # Token symbol missing
        with self.assertRaises(ValueError):
            tx = self.make_tx(symbol=b'')
            tx.sign(self.alice)

        # Token name missing
        with self.assertRaises(ValueError):
            tx = self.make_tx(name=b'')
            tx.sign(self.alice)

        # Empty initial_balances
        with self.assertRaises(ValueError):
            tx = self.make_tx(initial_balances=[])
            tx.sign(self.alice)

        # Invalid initial balances... 0!
        with self.assertRaises(ValueError):
            initial_balances_0_0 = [qrl_pb2.AddressAmount(address=self.alice.address, amount=0),
                                    qrl_pb2.AddressAmount(address=self.bob.address, amount=0)]
            tx = self.make_tx(initial_balances=initial_balances_0_0)
            tx.sign(self.alice)

        # Fee is -1
        with patch('qrl.core.txs.TokenTransaction.TokenTransaction.fee', new_callable=PropertyMock) as m_fee:
            m_fee.return_value = -1
            with self.assertRaises(ValueError):
                tx = self.make_tx()
                tx.sign(self.alice)

        # Invalid initial balances... -1!
        # tx = self.make_tx()
        # tx.sign(self.alice)
        # with patch('qrl.core.txs.TokenTransaction.TokenTransaction.initial_balances', new_callable=PropertyMock) as m_i_balances:
        #     m_i_balances.return_value = [-1, -1]
        #     with self.assertRaises(ValueError):
        #         tx.validate_or_raise()

    @patch('qrl.core.txs.Transaction.Transaction.validate_slave', return_value=True)
    def test_validate_extended(self, m_validate_slave, m_logger):
        """
        TokenTransaction.validate_extended checks for:
        1. valid master/slave
        2. from address is valid
        3. owner address is valid
        4. addresses that own the initial balances are valid
        5. that the AddressState has enough coins to pay the Transaction fee (because no coins are being transferred)
        6. OTS key reuse
        """
        tx = TokenTransaction.create(**self.params)

        m_addr_from_state = Mock(autospec=AddressState, name='addr_from State', balance=100)
        m_addr_from_pk_state = Mock(autospec=AddressState, name='addr_from_pk State')
        m_addr_from_pk_state.ots_key_reuse.return_value = False
        tx.sign(self.alice)
        result = tx.validate_extended(m_addr_from_state, m_addr_from_pk_state)
        self.assertTrue(result)

        m_validate_slave.return_value = False
        result = tx.validate_extended(m_addr_from_state, m_addr_from_pk_state)
        self.assertFalse(result)

        m_validate_slave.return_value = True
        with patch('qrl.core.txs.TokenTransaction.TokenTransaction.addr_from',
                   new_callable=PropertyMock) as m_addr_from:
            m_addr_from.return_value = b'Invalid Address'
            result = tx.validate_extended(m_addr_from_state, m_addr_from_pk_state)
            self.assertFalse(result)

        with patch('qrl.core.txs.TokenTransaction.TokenTransaction.owner', new_callable=PropertyMock) as m_owner:
            m_owner.return_value = b'Invalid Address'
            result = tx.validate_extended(m_addr_from_state, m_addr_from_pk_state)
            self.assertFalse(result)

        with patch('qrl.core.txs.TokenTransaction.TokenTransaction.initial_balances',
                   new_callable=PropertyMock) as m_address_balance:
            m_address_balance.return_value = [qrl_pb2.AddressAmount(address=b'Invalid Address 1', amount=1000),
                                              qrl_pb2.AddressAmount(address=b'Invalid Address 2', amount=1000)]
            result = tx.validate_extended(m_addr_from_state, m_addr_from_pk_state)
            self.assertFalse(result)

        m_addr_from_state.balance = 0
        result = tx.validate_extended(m_addr_from_state, m_addr_from_pk_state)
        self.assertFalse(result)
        m_addr_from_state.balance = 100

        m_addr_from_pk_state.ots_key_reuse.return_value = True
        result = tx.validate_extended(m_addr_from_state, m_addr_from_pk_state)
        self.assertFalse(result)

    def test_affected_address(self, m_logger):
        tx = TokenTransaction.create(**self.params)
        # Default params should result in 2 affected addresses
        result = set()
        tx.set_affected_address(result)
        self.assertEqual(2, len(result))

        # If the slave is a recipient of tokens, he should be included too.
        slave = get_slave_xmss()
        result = set()
        self.initial_balances_valid.append(qrl_pb2.AddressAmount(address=slave.address, amount=1000))
        tx = TokenTransaction.create(symbol=b'QRL',
                                     name=b'Quantum Resistant Ledger',
                                     owner=self.alice.address,
                                     decimals=15,
                                     initial_balances=self.initial_balances_valid,
                                     fee=1,
                                     xmss_pk=self.alice.pk)
        tx.set_affected_address(result)
        self.assertEqual(3, len(result))



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from grpc import StatusCode

from pyqrllib.pyqrllib import bin2hstr

from qrl.core import config
from qrl.core.qrlnode import QRLNode
from qrl.generated import qrlmining_pb2
from qrl.generated.qrlmining_pb2_grpc import MiningAPIServicer
from qrl.services.grpcHelper import GrpcExceptionWrapper


class MiningAPIService(MiningAPIServicer):
    MAX_REQUEST_QUANTITY = 100

    def __init__(self, qrlnode: QRLNode):
        self.qrlnode = qrlnode

    @GrpcExceptionWrapper(qrlmining_pb2.GetBlockMiningCompatibleResp, StatusCode.UNKNOWN)
    def GetBlockMiningCompatible(self,
                                 request: qrlmining_pb2.GetBlockMiningCompatibleReq,
                                 context) -> qrlmining_pb2.GetBlockMiningCompatibleResp:

        blockheader, block_metadata = self.qrlnode.get_blockheader_and_metadata(request.height)

        response = qrlmining_pb2.GetBlockMiningCompatibleResp()
        if blockheader is not None and block_metadata is not None:
            response = qrlmining_pb2.GetBlockMiningCompatibleResp(
                blockheader=blockheader.pbdata,
                blockmetadata=block_metadata.pbdata)

        return response

    @GrpcExceptionWrapper(qrlmining_pb2.GetLastBlockHeaderResp, StatusCode.UNKNOWN)
    def GetLastBlockHeader(self,
                           request: qrlmining_pb2.GetLastBlockHeaderReq,
                           context) -> qrlmining_pb2.GetLastBlockHeaderResp:
        response = qrlmining_pb2.GetLastBlockHeaderResp()

        blockheader, block_metadata = self.qrlnode.get_blockheader_and_metadata(request.height)

        response.difficulty = int(bin2hstr(block_metadata.block_difficulty), 16)
        response.height = self.qrlnode.block_height
        response.timestamp = blockheader.timestamp
        response.reward = blockheader.block_reward + blockheader.fee_reward
        response.hash = bin2hstr(blockheader.headerhash)
        response.depth = self.qrlnode.block_height - blockheader.block_number

        return response

    @GrpcExceptionWrapper(qrlmining_pb2.GetBlockToMineResp, StatusCode.UNKNOWN)
    def GetBlockToMine(self,
                       request: qrlmining_pb2.GetBlockToMineReq,
                       context) -> qrlmining_pb2.GetBlockToMineResp:

        response = qrlmining_pb2.GetBlockToMineResp()

        blocktemplate_blob_and_difficulty = self.qrlnode.get_block_to_mine(request.wallet_address)

        if blocktemplate_blob_and_difficulty:
            response.blocktemplate_blob = blocktemplate_blob_and_difficulty[0]
            response.difficulty = blocktemplate_blob_and_difficulty[1]
            response.height = self.qrlnode.block_height + 1
            response.reserved_offset = config.dev.extra_nonce_offset

        return response

    @GrpcExceptionWrapper(qrlmining_pb2.GetBlockToMineResp, StatusCode.UNKNOWN)
    def SubmitMinedBlock(self,
                         request: qrlmining_pb2.SubmitMinedBlockReq,
                         context) -> qrlmining_pb2.SubmitMinedBlockResp:
        response = qrlmining_pb2.SubmitMinedBlockResp()

        response.error = not self.qrlnode.submit_mined_block(request.blob)

        return response



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import functools
import threading

from pyqryptonight.pyqryptonight import PoWHelper

from qrl.core.Singleton import Singleton


class PoWValidator(object, metaclass=Singleton):
    def __init__(self):
        self.lock = threading.Lock()
        self._powv = PoWHelper()

    def verify_input(self, mining_blob, target):
        return self._verify_input_cached(mining_blob, target)

    @functools.lru_cache(maxsize=5)
    def _verify_input_cached(self, mining_blob, target):
        return self._powv.verifyInput(mining_blob, target)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from collections import Set

import simplejson as json

from qrl.core.misc import ntp, logger


class ExpiringSet(Set):
    def __init__(self, expiration_time, filename=None):
        super().__init__()
        self.expiration_time = expiration_time
        self._data = dict()
        self._filename = filename
        self._load()

    def __contains__(self, x: object) -> bool:
        self._refresh()
        return x in self._data

    def __len__(self) -> int:
        self._refresh()
        return len(self._data)

    def __iter__(self):
        self._refresh()
        return self._data.keys().__iter__()

    def add(self, x):
        current_time = ntp.getTime()
        self._data[x] = current_time + self.expiration_time
        self._store()

    def _refresh(self):
        # TODO: refactored from banned peers. Rework to use a priority queue instead
        current_time = ntp.getTime()

        len_before = len(self._data)
        self._data = {k: v for k, v in self._data.items() if v > current_time}
        len_after = len(self._data)

        # FIXME: Drop peers beyond configuration limit

        if len_before != len_after:
            self._store()

    def _store(self):
        if self._filename is not None:
            try:
                with open(self._filename, 'w') as f:
                    json.dump(self._data, f)
            except Exception as e:
                logger.error("not possible to save banned peers")
                logger.exception(e)

    def _load(self):
        if self._filename is not None:
            try:
                with open(self._filename, 'r') as f:
                    self._data = json.load(f)
                self._refresh()
            except FileNotFoundError:
                self._data = dict()



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import simplejson as json
import grpc
from google.protobuf.json_format import MessageToJson
from qrl.core import config
from qrl.core.AddressState import AddressState
from qrl.crypto.xmss import XMSS
from qrl.core.txs.Transaction import Transaction
from qrl.core.txs.TransferTransaction import TransferTransaction
from pyqrllib.pyqrllib import hstr2bin, bin2hstr

from qrl.generated import qrl_pb2_grpc, qrl_pb2, qrlmining_pb2, qrlmining_pb2_grpc
from flask import Flask, Response, request
from jsonrpc.backend.flask import api

app = Flask(__name__)


def read_slaves(slaves_filename):
    with open(slaves_filename, 'r') as f:
        slave_data = json.load(f)
        slave_data[0] = bytes(hstr2bin(slave_data[0]))
        return slave_data


def get_addr_state(addr: bytes) -> AddressState:
    stub = get_public_stub()
    response = stub.GetAddressState(request=qrl_pb2.GetAddressStateReq(address=addr))
    return AddressState(response.state)


def set_unused_ots_key(xmss, addr_state, start=0):
    for i in range(start, 2 ** xmss.height):
        if not addr_state.ots_key_reuse(i):
            xmss.set_ots_index(i)
            return True
    return False


def valid_payment_permission(public_stub, master_address_state, payment_xmss, json_slave_txn):
    access_type = master_address_state.get_slave_permission(payment_xmss.pk)

    if access_type == -1:
        tx = Transaction.from_json(json_slave_txn)
        public_stub.PushTransaction(request=qrl_pb2.PushTransactionReq(transaction_signed=tx.pbdata))
        return None

    if access_type == 0:
        return True

    return False


def get_unused_payment_xmss(public_stub):
    global payment_slaves
    global payment_xmss

    master_address = payment_slaves[0]
    master_address_state = get_addr_state(master_address)

    if payment_xmss:
        addr_state = get_addr_state(payment_xmss.address)
        if set_unused_ots_key(payment_xmss, addr_state, payment_xmss.ots_index):
            if valid_payment_permission(public_stub, master_address_state, payment_xmss, payment_slaves[2]):
                return payment_xmss
        else:
            payment_xmss = None

    if not payment_xmss:
        unused_ots_found = False
        for slave_seed in payment_slaves[1]:
            xmss = XMSS.from_extended_seed(slave_seed)
            addr_state = get_addr_state(xmss.address)
            if set_unused_ots_key(xmss, addr_state):  # Unused ots_key_found
                payment_xmss = xmss
                unused_ots_found = True
                break

        if not unused_ots_found:  # Unused ots_key_found
            return None

    if not valid_payment_permission(public_stub, master_address_state, payment_xmss, payment_slaves[2]):
        return None

    return payment_xmss


@app.route('/api/<api_method_name>')
def api_proxy(api_method_name):
    """
    Proxy JSON RPC requests to the gRPC server as well as converts back gRPC response
    to JSON.
    :param api_method_name:
    :return:
    """
    stub = qrl_pb2_grpc.PublicAPIStub(grpc.insecure_channel('{}:{}'.format(config.user.public_api_host,
                                                                           config.user.public_api_port)))
    public_api = qrl_pb2.DESCRIPTOR.services_by_name['PublicAPI']
    api_method = public_api.FindMethodByName(api_method_name)
    api_request = getattr(qrl_pb2, api_method.input_type.name)()

    for arg in request.args:
        if arg not in api_method.input_type.fields_by_name:
            raise Exception('Invalid args %s', arg)
        data_type = type(getattr(api_request, arg))
        if data_type == bool and request.args[arg].lower() == 'false':
            continue
        value = data_type(request.args.get(arg, type=data_type))
        setattr(api_request, arg, value)

    resp = getattr(stub, api_method_name)(api_request, timeout=10)
    return Response(response=MessageToJson(resp, sort_keys=True), status=200, mimetype='application/json')


def get_mining_stub():
    global mining_stub
    return mining_stub


def get_public_stub():
    global public_stub
    return public_stub


@api.dispatcher.add_method
def getlastblockheader(height=0):
    stub = get_mining_stub()
    request = qrlmining_pb2.GetLastBlockHeaderReq(height=height)
    grpc_response = stub.GetLastBlockHeader(request=request, timeout=10)

    block_header = {
        'difficulty': grpc_response.difficulty,
        'height': grpc_response.height,
        'timestamp': grpc_response.timestamp,
        'reward': grpc_response.reward,
        'hash': grpc_response.hash,
        'depth': grpc_response.depth
    }

    resp = {
        "block_header": block_header,
        "status": "OK"
    }
    return resp


@api.dispatcher.add_method
def getblockheaderbyheight(height):
    return getlastblockheader(height)


@api.dispatcher.add_method
def getblocktemplate(reserve_size, wallet_address):
    stub = get_mining_stub()
    request = qrlmining_pb2.GetBlockToMineReq(wallet_address=wallet_address.encode())
    grpc_response = stub.GetBlockToMine(request=request, timeout=10)
    resp = {
        'blocktemplate_blob': grpc_response.blocktemplate_blob,
        'difficulty': grpc_response.difficulty,
        'height': grpc_response.height,
        'reserved_offset': grpc_response.reserved_offset,
        'status': 'OK'
    }

    return resp


@api.dispatcher.add_method
def submitblock(blob):
    stub = get_mining_stub()
    request = qrlmining_pb2.SubmitMinedBlockReq(blob=bytes(hstr2bin(blob)))
    response = stub.SubmitMinedBlock(request=request, timeout=10)
    if response.error:
        raise Exception  # Mining pool expected exception when block submission fails
    return MessageToJson(response, sort_keys=True)


@api.dispatcher.add_method
def getblockminingcompatible(height):
    stub = get_mining_stub()
    request = qrlmining_pb2.GetBlockMiningCompatibleReq(height=height)
    response = stub.GetBlockMiningCompatible(request=request, timeout=10)
    return MessageToJson(response, sort_keys=True)


@api.dispatcher.add_method
def transfer(destinations, fee, mixin, unlock_time):
    if len(destinations) > config.dev.transaction_multi_output_limit:
        raise Exception('Payment Failed: Amount exceeds the allowed limit')

    addrs_to = []
    amounts = []

    for tx in destinations:
        addrs_to.append(bytes(hstr2bin(tx['address'][1:])))  # Skipping 'Q'
        amounts.append(tx['amount'])

    stub = get_public_stub()

    xmss = get_unused_payment_xmss(stub)
    if not xmss:
        raise Exception('Payment Failed: No Unused Payment XMSS found')

    tx = TransferTransaction.create(addrs_to=addrs_to,
                                    amounts=amounts,
                                    fee=fee,
                                    xmss_pk=xmss.pk,
                                    master_addr=payment_slaves[0])

    tx.sign(xmss)

    response = stub.PushTransaction(request=qrl_pb2.PushTransactionReq(transaction_signed=tx.pbdata))

    if response.error_code != 3:
        raise Exception('Transaction Submission Failed, Response Code: %s', response.error_code)

    response = {'tx_hash': bin2hstr(tx.txhash)}

    return response


app.add_url_rule('/json_rpc', 'api', api.as_view(), methods=['POST'])


def main():
    global payment_slaves, payment_xmss
    global mining_stub, public_stub
    mining_stub = qrlmining_pb2_grpc.MiningAPIStub(grpc.insecure_channel('{0}:{1}'.format(config.user.mining_api_host,
                                                                                          config.user.mining_api_port)))
    public_stub = qrl_pb2_grpc.PublicAPIStub(grpc.insecure_channel('{0}:{1}'.format(config.user.public_api_host,
                                                                                    config.user.public_api_port)))
    payment_xmss = None
    payment_slaves = read_slaves(config.user.mining_pool_payment_wallet_path)
    app.run(host=config.user.grpc_proxy_host, port=config.user.grpc_proxy_port)


if __name__ == '__main__':
    main()



from unittest import TestCase

import simplejson as json
from mock import patch, PropertyMock, Mock
from pyqrllib.pyqrllib import bin2hstr

from qrl.core.misc import logger
from qrl.core.AddressState import AddressState
from qrl.core.txs.Transaction import Transaction
from qrl.core.txs.TransferTokenTransaction import TransferTokenTransaction
from tests.core.txs.testdata import test_json_TransferToken, test_signature_TransferToken
from tests.misc.helper import get_alice_xmss, get_bob_xmss, get_slave_xmss

logger.initialize_default()


@patch('qrl.core.txs.Transaction.logger')
class TestTransferTokenTransaction(TestCase):

    def setUp(self):
        self.alice = get_alice_xmss()
        self.bob = get_bob_xmss()

        self.alice.set_ots_index(10)
        self.maxDiff = None

    def default_params(self):
        params = {
            "token_txhash": b'',
            "addrs_to": [self.bob.address],
            "amounts": [100],
            "fee": 1,
            "xmss_pk": self.alice.pk,
        }
        return params

    def test_create(self, m_logger):
        tx = TransferTokenTransaction.create(token_txhash=b'000000000000000',
                                             addrs_to=[self.bob.address],
                                             amounts=[200000],
                                             fee=1,
                                             xmss_pk=self.alice.pk)
        self.assertTrue(tx)

    def test_to_json(self, m_logger):
        tx = TransferTokenTransaction.create(token_txhash=b'000000000000000',
                                             addrs_to=[self.bob.address],
                                             amounts=[200000],
                                             fee=1,
                                             xmss_pk=self.alice.pk)
        txjson = tx.to_json()

        self.assertEqual(json.loads(test_json_TransferToken), json.loads(txjson))

    def test_from_json(self, m_logger):
        tx = Transaction.from_json(test_json_TransferToken)
        tx.sign(self.alice)

        self.assertIsInstance(tx, TransferTokenTransaction)

        # Test that common Transaction components were copied over.
        self.assertEqual('010300a1da274e68c88b0ccf448e0b1916fa789b01eb2ed4e9ad565ce264c9390782a9c61ac02f',
                         bin2hstr(tx.addr_from))
        self.assertEqual('01030038ea6375069f8272cc1a6601b3c76c21519455603d370036b97c779ada356'
                         '5854e3983bd564298c49ae2e7fa6e28d4b954d8cd59398f1225b08d6144854aee0e',
                         bin2hstr(tx.PK))
        self.assertEqual(b'000000000000000', tx.token_txhash)
        self.assertEqual(200000, tx.total_amount)
        self.assertEqual('390b159b34cffd29d4271a19679ff227df2ccd471078f177a7b58ca5f5d999f0', bin2hstr(tx.txhash))
        self.assertEqual(10, tx.ots_key)

        # z = bin2hstr(tx.signature)
        # print('"', end='')
        # for i in range(len(z)):
        #     print(z[i], end='')
        #     if (i + 1) % 64 == 0:
        #         print('" \\', end='')
        #         print('')
        #         print(' ' * len('test_signature_TransferToken = '), end='')
        #         print('"', end='')

        self.assertEqual(test_signature_TransferToken, bin2hstr(tx.signature))

        self.assertEqual(1, tx.fee)

    def test_validate_tx(self, m_logger):
        tx = TransferTokenTransaction.create(token_txhash=b'000000000000000',
                                             addrs_to=[self.bob.address],
                                             amounts=[200000],
                                             fee=1,
                                             xmss_pk=self.alice.pk)

        # We must sign the tx before validation will work.
        tx.sign(self.alice)

        # We have not touched the tx: validation should pass.
        self.assertTrue(tx.validate_or_raise())

    def test_validate_tx2(self, m_logger):
        params = self.default_params()
        tx = TransferTokenTransaction.create(**params)
        tx.sign(self.alice)

        self.assertTrue(tx.validate_or_raise())

        tx._data.transaction_hash = b'abc'

        # Should fail, as we have modified with invalid transaction_hash
        with self.assertRaises(ValueError):
            tx.validate_or_raise()

    def test_state_validate_tx_custom(self, m_logger):
        """
        TransferTokenTransaction._validate_custom() checks for:
        1. 0 amounts
        2. fee < 0
        3. multi-send: too many recipients
        4. multi-send: recipient addresses and amounts not the same length
        5. invalid addr_from
        6. invalid addr_to
        """
        slave = get_slave_xmss()
        params = self.default_params()
        tx = TransferTokenTransaction.create(**params)
        tx.sign(self.alice)

        result = tx.validate_or_raise()
        self.assertTrue(result)

        params = self.default_params()
        params["addrs_to"] = [self.bob.address, slave.address]
        params["amounts"] = [1, 0]
        with self.assertRaises(ValueError):
            tx = TransferTokenTransaction.create(**params)

        # Protobuf validation doesn't allow negative fees already
        params = self.default_params()
        tx = TransferTokenTransaction.create(**params)
        tx.sign(self.alice)
        with patch('qrl.core.txs.Transaction.Transaction.fee', new_callable=PropertyMock) as m_fee:
            m_fee.return_value = -1
            with self.assertRaises(ValueError):
                tx.validate_or_raise()

        params = self.default_params()
        params["addrs_to"] = [self.bob.address, slave.address, self.alice.address]
        params["amounts"] = [2, 3, 5]
        tx = TransferTokenTransaction.create(**params)
        tx.sign(self.alice)
        with patch('qrl.core.txs.TransferTokenTransaction.config', autospec=True) as m_config:
            m_config.dev.transaction_multi_output_limit = 1
            with self.assertRaises(ValueError):
                tx.validate_or_raise()

        # TX signing already fails if addrs_to and amounts are unequal length
        params = self.default_params()
        tx = TransferTokenTransaction.create(**params)
        tx.sign(self.alice)
        with patch('qrl.core.txs.TransferTokenTransaction.TransferTokenTransaction.addrs_to',
                   new_callable=PropertyMock) as m_addrs_to:
            m_addrs_to.return_value = [self.bob.address, slave.address]
            with self.assertRaises(ValueError):
                tx.validate_or_raise()

        params = self.default_params()
        params["master_addr"] = b'Bad QRL Address'
        with self.assertRaises(ValueError):
            tx = TransferTokenTransaction.create(**params)

        params = self.default_params()
        params["addrs_to"] = [self.bob.address, b'Bad QRL address']
        params["amounts"] = [100, 200]
        with self.assertRaises(ValueError):
            tx = TransferTokenTransaction.create(**params)

    @patch('qrl.core.txs.Transaction.Transaction.validate_slave', return_value=True)
    def test_validate_extended(self, m_validate_slave, m_logger):
        """
        TransferTokenTransaction.validate_extended checks for:
        1. valid master/slave
        2. negative fee, negative total token amounts transferred
        3. addr_from has enough funds for the fee
        4. if addr_from owns any tokens to begin with
        5. if addr_from has enough tokens
        6. addr_from ots_key reuse
        """
        m_addr_from_state = Mock(autospec=AddressState, name='addr_from State', balance=100)
        m_addr_from_state.is_token_exists.return_value = True
        m_addr_from_state.get_token_balance.return_value = 1000

        m_addr_from_pk_state = Mock(autospec=AddressState, name='addr_from_pk State')
        m_addr_from_pk_state.ots_key_reuse.return_value = False

        params = self.default_params()
        tx = TransferTokenTransaction.create(**params)
        tx.sign(self.alice)

        result = tx.validate_extended(m_addr_from_state, m_addr_from_pk_state)
        self.assertTrue(result)

        # Invalid master XMSS/slave XMSS relationship
        m_validate_slave.return_value = False
        result = tx.validate_extended(m_addr_from_state, m_addr_from_pk_state)
        self.assertFalse(result)
        m_validate_slave.return_value = True

        # fee = -1
        with patch('qrl.core.txs.TransferTokenTransaction.TransferTokenTransaction.fee',
                   new_callable=PropertyMock) as m_fee:
            m_fee.return_value = -1
            result = tx.validate_extended(m_addr_from_state, m_addr_from_pk_state)
            self.assertFalse(result)

        # total_amount = -1
        with patch('qrl.core.txs.TransferTokenTransaction.TransferTokenTransaction.total_amount',
                   new_callable=PropertyMock) as m_total_amount:
            m_total_amount.return_value = -100
            result = tx.validate_extended(m_addr_from_state, m_addr_from_pk_state)
            self.assertFalse(result)

        # balance = 0, cannot pay the Transaction fee
        m_addr_from_state.balance = 0
        result = tx.validate_extended(m_addr_from_state, m_addr_from_pk_state)
        self.assertFalse(result)
        m_addr_from_state.balance = 100

        # addr_from doesn't have these tokens
        m_addr_from_state.is_token_exists.return_value = False
        result = tx.validate_extended(m_addr_from_state, m_addr_from_pk_state)
        self.assertFalse(result)
        m_addr_from_state.is_token_exists.return_value = True

        # addr_from doesn't have enough tokens
        m_addr_from_state.get_token_balance.return_value = 99
        result = tx.validate_extended(m_addr_from_state, m_addr_from_pk_state)
        self.assertFalse(result)
        m_addr_from_state.get_token_balance.return_value = 1000

        # addr_from_pk has used this OTS key before
        m_addr_from_pk_state.ots_key_reuse.return_value = True
        result = tx.validate_extended(m_addr_from_state, m_addr_from_pk_state)
        self.assertFalse(result)

    def test_set_affected_address(self, m_logger):
        result = set()
        params = self.default_params()
        tx = TransferTokenTransaction.create(**params)
        tx.set_affected_address(result)
        self.assertEqual(2, len(result))

        params = self.default_params()
        params["addrs_to"] = [self.bob.address, get_slave_xmss().address]
        params["amounts"] = [100, 200]
        tx = TransferTokenTransaction.create(**params)
        tx.set_affected_address(result)
        self.assertEqual(3, len(result))



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import os
import time
from unittest import TestCase

from mock import Mock, patch, mock
from pyqrllib.pyqrllib import hstr2bin
from pyqryptonight.pyqryptonight import StringToUInt256

from qrl.core import config
from qrl.core.misc import logger
from qrl.core.p2p.IPMetadata import IPMetadata
from qrl.core.p2p.p2pPeerManager import P2PPeerManager
from qrl.core.p2p.p2pfactory import P2PFactory
from qrl.core.p2p.p2pprotocol import P2PProtocol
from qrl.generated import qrl_pb2, qrllegacy_pb2
from tests.misc.helper import replacement_getTime
from tests.misc.helper import set_qrl_dir

logger.initialize_default()


def make_channel(name=''):
    channel = Mock(autospec=P2PProtocol, name=name)
    channel.factory = Mock(autospec=P2PFactory)
    return channel


def make_node_chain_state():
    node_chain_state = qrl_pb2.NodeChainState(block_number=0,
                                              header_hash=b'',
                                              cumulative_difficulty=b'0' * 32,
                                              timestamp=int(time.time()))
    return node_chain_state


# Some functions have logger patched out, so that tests are not too noisy when unexpected things happen.
@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
class TestP2PPeerManager(TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.peer_manager = P2PPeerManager()

    def tearDown(self):
        del self.peer_manager

    def populate_peer_manager(self):
        channel_1 = make_channel('channel_1')
        channel_2 = make_channel('channel_2')
        channel_3 = make_channel('channel_3')

        self.peer_manager._channels = [channel_1, channel_2, channel_3]

        self.peer_manager._peer_node_status[channel_1] = make_node_chain_state()
        self.peer_manager._peer_node_status[channel_2] = make_node_chain_state()
        self.peer_manager._peer_node_status[channel_3] = make_node_chain_state()

        return channel_1, channel_2, channel_3

    def test_load_peer_addresses_peer_qrl(self):
        """
        Should load peers from peers.qrl AND from config.user.peer_list
        """
        with set_qrl_dir('peers') as tmp_dir:
            self.peer_manager.peers_path = os.path.join(tmp_dir, 'peers.json')

            self.peer_manager.load_peer_addresses()

            # This was in peers.qrl
            self.assertIn(IPMetadata.canonical_full_address('34.208.138.15'), self.peer_manager.known_peer_addresses)
            # config.user.peer_list is all in there too
            for p in config.user.peer_list:
                self.assertIn(IPMetadata.canonical_full_address(p), self.peer_manager.known_peer_addresses)

    def test_load_peer_addresses_no_file(self):
        """
        If no peers.qrl exists, use config.user.peer_list
        """
        with set_qrl_dir('no_data') as tmp_dir:
            self.peer_manager.peers_path = os.path.join(tmp_dir, config.dev.peers_filename)

            self.peer_manager.load_peer_addresses()

            # config.user.peer_list is all in there too
            self.assertEqual(len(config.user.peer_list), len(self.peer_manager.known_peer_addresses))
            for p in config.user.peer_list:
                self.assertIn(IPMetadata.canonical_full_address(p), self.peer_manager.known_peer_addresses)

    @patch('qrl.core.p2p.p2pPeerManager.logger', autospec=True)
    def test_load_peer_addresses_corrupt_file(self, logger):
        """
        If peers.qrl is corrupt, load_peer_addresses should rewrite it with valid peers from config.user.peer_list
        """
        with set_qrl_dir('peers') as tmp_dir:
            self.peer_manager.peers_path = os.path.join(tmp_dir, 'peers_corrupt.json')

            self.peer_manager.load_peer_addresses()

            # config.user.peer_list is all in there too
            with open(os.path.join(tmp_dir, 'peers_corrupt.json')) as f:
                contents = f.read()
                for p in config.user.peer_list:
                    self.assertIn(p, contents)

    @patch('qrl.core.p2p.p2pPeerManager.logger', autospec=True)
    def test_combine_peer_lists_works(self, logger):
        """
        combine_peer_lists takes: a set of IP:PORTs from a peer; the peer's ip; the peer's port
        It validates the set of IP:PORTs and adds valid ones to the node's peerlist.
        Global IP addresses only.
        Because we are connected to the peer from whom we downloaded this list, the peer's ip and port are validated
        differently and added to the peerlist as well.
        If the same IP appears with different ports, they should appear as different records in the peerlist (max 2)
        """
        # The second IP in the set has an invalid port, but the peer to which we are connected should be in the set.
        result = self.peer_manager.combine_peer_lists({'1.1.1.1:9000', '1.2.3.4:65536'},
                                                      ['187.0.0.1:1000'], check_global=True)
        self.assertEqual(result, {'1.1.1.1:9000', '187.0.0.1:1000'})

        # What happens if the peer we are connected to now has a different port? it should replace the entry in the set.
        result = self.peer_manager.combine_peer_lists({'1.1.1.1:9000', '127.0.0.1:1000'},
                                                      ['187.0.0.1:2000'], check_global=True)
        self.assertEqual(result, {'1.1.1.1:9000', '187.0.0.1:2000'})

    @patch('qrl.core.p2p.p2pPeerManager.logger', autospec=True)
    def test_combine_peer_lists_bad_ip_list(self, logger):
        """
        combine_peer_lists should revalidate all existing ip:port pairs when adding a new one to the list.
        but why isn't validation happening in another layer so that the set is always clean?
        """
        bad_ip_list_result = self.peer_manager.combine_peer_lists({'256.256.256.256:9000', '187.0.0.3:90000'},
                                                                  ['187.0.0.1:9000'], check_global=True)
        self.assertEqual(bad_ip_list_result, {'187.0.0.1:9000'})

        # A local IP in the incoming set should be ignored.
        result = self.peer_manager.combine_peer_lists({'1.1.1.1:8000', '127.0.0.1:1111'},
                                                      ['127.0.0.1:3000'], check_global=True)
        self.assertEqual(result, {'1.1.1.1:8000'})

        # A bad IP in the set should not pass.
        result = self.peer_manager.combine_peer_lists({'255.255.255.255:8000'},
                                                      ['127.0.0.1:3000'], check_global=True)
        self.assertEqual(result, set())

    @patch('qrl.core.p2p.p2pPeerManager.logger', autospec=True)
    def test_update_peer_addresses_also_connects_to_new_peers(self, logger):
        """
        extend_known_peers() not only writes out, it automatically connects to any new peers.
        """
        self.peer_manager._p2pfactory = Mock()
        self.peer_manager._known_peers = {'1.1.1.1:9000'}
        with set_qrl_dir('no_data') as tempdir:
            self.peer_manager.peers_path = os.path.join(tempdir, config.dev.peers_filename)
            self.peer_manager.extend_known_peers({'2.2.2.2:9000'})

        self.peer_manager._p2pfactory.connect_peer.assert_called_once_with('2.2.2.2:9000')

    def test_remove_channel(self):
        """
        For each channel in _channels, there should be a corresponding entry in _peer_node_status.
        remove_channel() makes sure there is a 1:1 correspondence.

        It also shouldn't make trouble if the _peer_node_status is missing.
        """
        channel_2, channel_3 = self.populate_peer_manager()[1:3]

        self.peer_manager.remove_channel(channel_3)
        self.assertEqual(len(self.peer_manager._channels), 2)
        self.assertEqual(len(self.peer_manager._peer_node_status), 2)

        # It also shouldn't make trouble if the _peer_node_status is missing.
        del self.peer_manager._peer_node_status[channel_2]
        self.peer_manager.remove_channel(channel_2)
        self.assertEqual(len(self.peer_manager._channels), 1)
        self.assertEqual(len(self.peer_manager._peer_node_status), 1)

    def test_new_channel(self):
        """
        new_channel() makes sure that for each channel P2PPeerManager has in _channels, there is a corresponding
        record in _peer_node_status.
        It also tells the channel to call P2PPeerManager's handle_* functions whenever certain messages come in.
        """
        channel = make_channel()
        self.peer_manager.new_channel(channel)
        self.assertEqual(len(self.peer_manager._channels), len(self.peer_manager._peer_node_status), 1)

        self.assertEqual(channel.register.call_count, 5)

    def test_handle_version(self):
        """
        When a version message arrives from a peer, and all else is normal:
        A version request message was not sent to the peer.
        The peer is not banned.
        P2PProtocol.loseConnection() is not called.
        """
        channel = make_channel()

        message = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.VE,
                                              veData=qrllegacy_pb2.VEData(version=config.dev.version,
                                                                          genesis_prev_hash=config.user.genesis_prev_headerhash,
                                                                          rate_limit=config.user.peer_rate_limit))
        self.peer_manager.handle_version(channel, message)
        channel.peer_manager.ban_channel.assert_not_called()
        channel.loseConnection.assert_not_called()

    def test_handle_version_empty_version_message(self):
        """
        If the incoming version message has an empty version field, then send another version request.
        That message should have the node's version in it.
        """
        channel = make_channel()

        message = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.VE,
                                              veData=qrllegacy_pb2.VEData(version='',
                                                                          genesis_prev_hash=config.user.genesis_prev_headerhash,
                                                                          rate_limit=config.user.peer_rate_limit))
        self.peer_manager.handle_version(channel, message)
        self.assertEqual(channel.send.call_args[0][0].veData.version, config.dev.version)

    def test_handle_version_wrong_genesis_prev_headerhash(self):
        """
        If the genesis_prev_headerhash is different, the nodes should disconnect from each other.
        """
        channel = make_channel()

        message = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.VE,
                                              veData=qrllegacy_pb2.VEData(version=config.dev.version,
                                                                          genesis_prev_hash=b'TEST123',
                                                                          rate_limit=config.user.peer_rate_limit))
        self.peer_manager.handle_version(channel, message)
        channel.loseConnection.assert_any_call()

    def test_get_better_difficulty(self):
        """
        Returns the P2PProtocol that has a higher cumulative difficulty than our node.
        Should not take any other variable into account,including block_number or timestamp.
        Also updates the 'best seen cumulative difficulty' variable.
        """
        test_cumulative_difficulty = StringToUInt256('5000')

        channel_1, channel_2, channel_3 = self.populate_peer_manager()

        self.peer_manager._peer_node_status[channel_1].cumulative_difficulty = bytes(StringToUInt256('5001'))
        self.peer_manager._peer_node_status[channel_2].cumulative_difficulty = bytes(StringToUInt256('0'))
        self.peer_manager._peer_node_status[channel_3].cumulative_difficulty = bytes(StringToUInt256('3'))

        self.assertEqual(self.peer_manager.get_better_difficulty(test_cumulative_difficulty), channel_1)

        # But channel_1 has a very old timestamp! doesn't matter.
        self.peer_manager._peer_node_status[channel_1].timestamp = 1400000000
        self.assertEqual(self.peer_manager.get_better_difficulty(test_cumulative_difficulty), channel_1)

        # But other nodes have higher blockheight! Doesn't matter
        self.peer_manager._peer_node_status[channel_1].block_number = 1
        self.peer_manager._peer_node_status[channel_2].block_number = 4
        self.peer_manager._peer_node_status[channel_3].block_number = 5

        self.assertEqual(self.peer_manager.get_better_difficulty(test_cumulative_difficulty), channel_1)

    def test_get_better_difficulty_self(self):
        """
        If we have the highest cum-difficulty, this function should return None.
        """
        test_cumulative_difficulty = StringToUInt256('5000')
        channel_1, channel_2, channel_3 = self.populate_peer_manager()

        self.peer_manager._peer_node_status[channel_1].cumulative_difficulty = bytes(StringToUInt256('49'))
        self.peer_manager._peer_node_status[channel_2].cumulative_difficulty = bytes(StringToUInt256('4800'))
        self.peer_manager._peer_node_status[channel_3].cumulative_difficulty = bytes(StringToUInt256('3'))

        self.assertIsNone(self.peer_manager.get_better_difficulty(test_cumulative_difficulty))

    @patch('qrl.core.p2p.p2pPeerManager.logger', autospec=True)
    def test_handle_peer_list_works(self, logger):
        """
        Heavy error testing should be done in combine_peer_lists() and extend_known_peers(), which this fx uses.
        """
        peer_list_message = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.PL,
                                                        plData=qrllegacy_pb2.PLData(
                                                            peer_ips={'127.0.0.3:5000', '127.0.0.4:5001'},
                                                            public_port=9000))
        channel = make_channel()
        channel.host = IPMetadata('187.0.0.1', 9000)
        channel.peer = IPMetadata('187.0.0.2', 9000)

        # handle_peer_list() will call extend_known_peers(), so we gotta mock it out. It's tested elsewhere anyway.
        self.peer_manager.extend_known_peers = Mock(autospec=P2PPeerManager.extend_known_peers)
        self.peer_manager.handle_peer_list(channel, peer_list_message)
        self.peer_manager.extend_known_peers.assert_called_once_with({channel.peer.full_address})

    @patch('qrl.core.p2p.p2pPeerManager.logger', autospec=True)
    def test_handle_peer_list_empty_peer_list_message(self, logger):
        """
        An empty plData should result in no further processing being done.
        """
        peer_list_message = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.PL,
                                                        plData=qrllegacy_pb2.PLData(peer_ips={}, public_port=9000))
        channel = make_channel()
        channel.host_ip = '127.0.0.1'
        channel.peer_ip = '127.0.0.2'

        self.peer_manager.extend_known_peers = Mock(autospec=P2PPeerManager.extend_known_peers)
        self.peer_manager.combine_peer_lists = Mock(autospec=P2PPeerManager.combine_peer_lists)

        self.peer_manager.handle_peer_list(channel, peer_list_message)

        self.peer_manager.combine_peer_lists.assert_not_called()
        self.peer_manager.extend_known_peers.assert_not_called()

    @patch('qrl.core.p2p.p2pPeerManager.logger', autospec=True)
    @patch('qrl.core.p2p.p2pPeerManager.config', autospec=True)
    def test_handle_peer_list_peer_discovery_disabled(self, config, logger):
        """
        If enable_peer_discovery is False, no further processing should be done.
        """
        config.user.enable_peer_discovery = False
        peer_list_message = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.PL,
                                                        plData=qrllegacy_pb2.PLData(
                                                            peer_ips={'127.0.0.3:5000', '127.0.0.4:5001'},
                                                            public_port=9000))
        channel = make_channel()
        channel.host_ip = '127.0.0.1'
        channel.peer_ip = '127.0.0.2'

        self.peer_manager.extend_known_peers = Mock(autospec=P2PPeerManager.extend_known_peers)
        self.peer_manager.combine_peer_lists = Mock(autospec=P2PPeerManager.combine_peer_lists)

        self.peer_manager.handle_peer_list(channel, peer_list_message)

        self.peer_manager.combine_peer_lists.assert_not_called()
        self.peer_manager.extend_known_peers.assert_not_called()

    def test_handle_sync(self):
        """
        If a message comes in saying 'Synced': the peer is synced.
        """
        sync_message = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.SYNC,
                                                   syncData=qrllegacy_pb2.SYNCData(state='Synced'))
        channel = make_channel()

        self.peer_manager.handle_sync(channel, sync_message)

    def test_handle_sync_unsynced(self):
        """
        If the message says anything else: the peer is unsynced.
        """
        sync_message = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.SYNC,
                                                   syncData=qrllegacy_pb2.SYNCData(state='Unsynced'))
        channel = make_channel()

        self.peer_manager.handle_sync(channel, sync_message)

    def test_handle_sync_blank(self):
        """
        If the message says '': the peer doesn't know, and isn't synced. We tell it that we are synced.
        """
        sync_message = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.SYNC,
                                                   syncData=qrllegacy_pb2.SYNCData(state=''))
        channel = make_channel()

        # But if we ourselves aren't synced, then we cannot say if anybody else is synced.
        channel.factory.synced = False
        self.peer_manager.handle_sync(channel, sync_message)
        channel.send_sync.assert_not_called()

        # If we are synced, then we can tell other nodes we are synced.
        channel.factory.synced = True
        self.peer_manager.handle_sync(channel, sync_message)
        channel.send_sync.assert_called_once_with(synced=True)

    def test_monitor_chain_state_works(self):
        """
        This function simply goes through each channel in P2PPeerManager and make sure the info it has on that
        is fresh. After a certain time period, the information is considered stale and the channel will be disconnected.
        If it doesn't have info on a particular P2PProtocol, then it will disconnect the channel
        (presumably the information should already have been exchanged when the channel was first connected)
        """
        channel_1, channel_2, channel_3 = self.populate_peer_manager()

        self.peer_manager.monitor_chain_state()

        channel_1.loseConnection.assert_not_called()
        channel_2.loseConnection.assert_not_called()
        channel_3.loseConnection.assert_not_called()

    def test_monitor_chain_state_stale_info(self):
        """
        After a certain time period, the information is considered stale and the channel will be disconnected.
        """
        channel_1 = self.populate_peer_manager()[0]
        self.peer_manager._peer_node_status[channel_1].timestamp = int(
            time.time()) - config.user.chain_state_timeout - 1

        self.peer_manager.monitor_chain_state()

        channel_1.loseConnection.assert_called_once_with()

    def test_monitor_chain_state_extra_channel(self):
        """
        If monitor_chain-state() doesn't have info on a particular P2PProtocol, then it will disconnect the channel
        new_channel() should have put the info there into _peer_node_status after all.
        """
        channel_3 = self.populate_peer_manager()[2]

        del self.peer_manager._peer_node_status[channel_3]

        self.peer_manager.monitor_chain_state()

        channel_3.loseConnection.assert_called_once_with()

    def test_handle_chain_state_works(self):
        """
        When a peer reports its chain state, we update the timestamp to reflect the time we received the message.
        Then we update our _peer_node_status.
        """
        chain_state_data = make_node_chain_state()

        chain_state_message = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.CHAINSTATE,
                                                          chainStateData=chain_state_data)
        channel = make_channel()

        # The P2PPeerManager should update its hash table with the info from chain_state_data.
        # Unfortunately it also updates the timestamp so we cannot simply compare the objects,
        # we have to compare the fields in the object excluding the timestamp.
        self.peer_manager.handle_chain_state(channel, chain_state_message)

        self.assertEqual(self.peer_manager._peer_node_status[channel].block_number, chain_state_data.block_number)
        self.assertEqual(self.peer_manager._peer_node_status[channel].header_hash, chain_state_data.header_hash)
        self.assertEqual(self.peer_manager._peer_node_status[channel].cumulative_difficulty,
                         chain_state_data.cumulative_difficulty)

    def test_handle_p2p_acknowledgement(self):
        """
        Once an acknowledgement is received from a peer, we can update rate counters and send the next message.
        """
        ack = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.P2P_ACK,
                                          p2pAckData=qrl_pb2.P2PAcknowledgement(bytes_processed=15))
        channel = make_channel()
        channel.bytes_sent = 20
        self.peer_manager.handle_p2p_acknowledgement(channel, ack)

        channel.send_next.assert_called_once_with()

    def test_handle_p2p_acknowledgement_negative_bytes_processed(self):
        """
        If we sent 10 bytes to the peer, but the peer says it processed 20 bytes, disconnect the peer.
        """
        ack = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.P2P_ACK,
                                          p2pAckData=qrl_pb2.P2PAcknowledgement(bytes_processed=20))
        channel = make_channel()
        channel.bytes_sent = 10
        self.peer_manager.handle_p2p_acknowledgement(channel, ack)

        channel.loseConnection.assert_called_once_with()

    @patch('qrl.core.p2p.p2pprotocol.P2PProtocol.peer')
    @patch('qrl.core.p2p.p2pprotocol.P2PProtocol.send')
    def test_trusted_message_count(self, send, get_peer):
        channel = P2PProtocol()
        with mock.patch('qrl.core.misc.ntp.getTime') as time_mock:
            time_mock.return_value = channel.connected_at + 100
            get_peer.return_value = IPMetadata('192.168.0.1', 1000)

            self.assertFalse(self.peer_manager.trusted_peer(channel))

            for _ in range(config.dev.trust_min_msgcount - 1):
                buffer = bytes(hstr2bin('000000191a170a0776657273696f6e120c67656e657369735f68617368'))
                channel.dataReceived(buffer)
                self.assertFalse(self.peer_manager.trusted_peer(channel))

            buffer = bytes(hstr2bin('000000191a170a0776657273696f6e120c67656e657369735f68617368'))
            channel.dataReceived(buffer)
            self.assertTrue(self.peer_manager.trusted_peer(channel))

    @patch('qrl.core.p2p.p2pprotocol.P2PProtocol.peer')
    @patch('qrl.core.p2p.p2pprotocol.P2PProtocol.send')
    def test_trusted_time(self, send, get_peer):
        channel = P2PProtocol()
        with mock.patch('qrl.core.misc.ntp.getTime') as time_mock:
            time_mock.return_value = channel.connected_at + 1
            get_peer.return_value = IPMetadata('192.168.0.1', 1000)

            for _ in range(config.dev.trust_min_msgcount):
                buffer = bytes(hstr2bin('000000191a170a0776657273696f6e120c67656e657369735f68617368'))
                channel.dataReceived(buffer)

            time_mock.return_value = channel.connected_at + config.dev.trust_min_conntime + 1
            self.assertTrue(self.peer_manager.trusted_peer(channel))



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase

from pyqrllib.pyqrllib import str2bin, XmssFast, bin2hstr, SHAKE_128, SHAKE_256, SHA2_256

from qrl.core.misc import logger
from qrl.crypto.xmss import XMSS
from tests.misc.helper import get_alice_xmss

logger.initialize_default()


class TestXMSS(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestXMSS, self).__init__(*args, **kwargs)

    def test_sign_verify(self):
        message = "This is a test"
        message_bin = str2bin(message)

        xmss_height = 10
        seed = bytearray([i for i in range(48)])
        xmss = XMSS(XmssFast(seed, xmss_height))

        pk = xmss.pk

        xmss.set_ots_index(1)

        for i in range(10):
            self.assertTrue(xmss.ots_index == i + 1)
            signature = xmss.sign(message_bin)
            self.assertTrue(XmssFast.verify(message_bin, signature, pk))

    def test_PK(self):
        xmss_height = 10
        seed = bytearray([i for i in range(48)])
        xmss = XMSS(XmssFast(seed, xmss_height))

        pk = xmss.pk
        self.assertEqual('010500ffc6e502e2a8244aed6a8cd67531e79f95baa638615ba789c194a1d15d7eb'
                         '77e4e3983bd564298c49ae2e7fa6e28d4b954d8cd59398f1225b08d6144854aee0e', bin2hstr(pk))

    def test_hash_function(self):
        xmss_height = 4
        seed = bytearray([i for i in range(48)])
        xmss = XMSS(XmssFast(seed, xmss_height, SHAKE_128))
        self.assertEqual('shake128', xmss.hash_function)

        xmss = XMSS(XmssFast(seed, xmss_height, SHAKE_256))
        self.assertEqual('shake256', xmss.hash_function)

        xmss = XMSS(XmssFast(seed, xmss_height, SHA2_256))
        self.assertEqual('sha2_256', xmss.hash_function)

    def test_signature_type(self):
        xmss_height = 4
        seed = bytearray([i for i in range(48)])
        xmss = XMSS(XmssFast(seed, xmss_height))
        self.assertEqual(0, xmss.signature_type)

    def test_from_height_custom_hash(self):
        xmss_height = 4
        xmss = XMSS.from_height(xmss_height, "shake128")
        self.assertEqual('shake128', xmss.hash_function)

    def test_get_height_from_sig_size(self):
        with self.assertRaises(Exception):
            XMSS.get_height_from_sig_size(2179)

        with self.assertRaises(Exception):
            XMSS.get_height_from_sig_size(0)

        with self.assertRaises(Exception):
            XMSS.get_height_from_sig_size(-1)

        height = XMSS.get_height_from_sig_size(3204)
        self.assertEqual(height, 32)

        height = XMSS.get_height_from_sig_size(2180)
        self.assertEqual(height, 0)

    def test_validate_signature(self):
        xmss = get_alice_xmss()
        xmss2 = get_alice_xmss(8)
        pk = xmss.pk
        signature = xmss.sign(b"hello")

        self.assertTrue(XMSS.validate_signature(signature, pk))

        with self.assertRaises(ValueError):
            XMSS.validate_signature(signature, None)

        self.assertFalse(XMSS.validate_signature(signature, xmss2.pk))



# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase

from qrl.core.misc import logger
from pyqryptonight.pyqryptonight import Qryptonight, UInt256ToString

logger.initialize_default()


class TestQryptonight(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestQryptonight, self).__init__(*args, **kwargs)

    def test_init(self):
        qn = Qryptonight()
        self.assertIsNotNone(qn)

    def test_hash(self):
        qn = Qryptonight()

        input = [0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07, 0x09,
                 0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07, 0x09,
                 0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07]
        output = qn.hash(input)

        output_expected = (
            0x96, 0x76, 0x53, 0x26, 0x2F, 0x9B, 0x15, 0x90,
            0xB9, 0x88, 0x0F, 0x64, 0xA3, 0x80, 0x8C, 0x4B,
            0x01, 0xEA, 0x29, 0x2C, 0x48, 0xFC, 0x7C, 0x47,
            0x0D, 0x25, 0x50, 0x00, 0x57, 0xCA, 0x07, 0x70,
        )

        self.assertEqual(output_expected, output)

    def test_hash2(self):
        qn = Qryptonight()

        input = [0x03, 0x05, 0x07, 0x09] * 30
        output = qn.hash(input)

        output_expected = (255, 253, 83, 78, 29, 24, 160, 224, 30, 240, 158,
                           39, 233, 125, 90, 170, 78, 59, 157, 146, 97, 86,
                           205, 161, 160, 155, 48, 144, 51, 148, 155, 99)

        self.assertEqual(output_expected, output)

    def test_hash3(self):
        qn = Qryptonight()

        input = [0x03, 0x05, 0x07, 0x09] * 200
        output = qn.hash(input)

        output_expected = (216, 31, 227, 138, 9, 118, 5, 200, 136,
                           40, 156, 168, 86, 35, 146, 223, 199, 76,
                           188, 213, 25, 117, 247, 195, 15, 183, 236,
                           219, 104, 212, 75, 211)

        self.assertEqual(output_expected, output)

    def test_empty(self):
        with self.assertRaises(TypeError):
            UInt256ToString(None)
        with self.assertRaises(ValueError):
            UInt256ToString(b'')



from unittest import TestCase

from mock import patch, Mock

from qrl.core.misc import logger
from qrl.core.AddressState import AddressState
from qrl.core.ChainManager import ChainManager
from qrl.core.txs.TransferTokenTransaction import TransferTokenTransaction
from tests.misc.helper import get_alice_xmss, get_bob_xmss, get_slave_xmss

logger.initialize_default()


@patch('qrl.core.txs.Transaction.Transaction._revert_state_changes_for_PK')
@patch('qrl.core.txs.Transaction.Transaction._apply_state_changes_for_PK')
@patch('qrl.core.txs.Transaction.logger')
class TestTransferTokenTransactionStateChanges(TestCase):
    def setUp(self):
        self.alice = get_alice_xmss()
        self.bob = get_bob_xmss()

        self.params = {
            "token_txhash": b'I declare the TEST token',
            "addrs_to": [self.bob.address],
            "amounts": [100],
            "fee": 1,
            "xmss_pk": self.alice.pk
        }

        self.unused_chain_manager_mock = Mock(autospec=ChainManager, name='unused ChainManager')

    def generate_addresses_state(self, tx):
        addresses_state = {
            self.alice.address: Mock(autospec=AddressState, name='alice AddressState',
                                     tokens={self.params["token_txhash"]: 1000}, transaction_hashes=[],
                                     balance=100),
            self.bob.address: Mock(autospec=AddressState, name='bob AddressState',
                                   tokens={self.params["token_txhash"]: 0},
                                   transaction_hashes=[],
                                   balance=0),
        }
        return addresses_state

    def test_apply_state_changes(self, m_logger, m_apply_state_PK, m_revert_state_PK):
        """
        Alice has 1000 tokens and 100 QRL, Bob has none. Alice sends some tokens to Bob.
        """
        tx = TransferTokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = self.generate_addresses_state(tx)
        tx.apply_state_changes(addresses_state)

        self.assertEqual(addresses_state[self.alice.address].balance, 99)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)
        addresses_state[self.alice.address].update_token_balance.assert_called_with(b'I declare the TEST token', -100)
        addresses_state[self.bob.address].update_token_balance.assert_called_with(b'I declare the TEST token', 100)
        self.assertEqual([tx.txhash], addresses_state[self.alice.address].transaction_hashes)
        self.assertEqual([tx.txhash], addresses_state[self.bob.address].transaction_hashes)

        m_apply_state_PK.assert_called_once()

    def test_apply_state_changes_multi_send(self, m_logger, m_apply_state_PK, m_revert_state_PK):
        """
        Alice has 1000 tokens and 100 QRL, Bob and Slave have none. Alice sends some tokens to Bob and Slave.
        """
        slave = get_slave_xmss()
        params = self.params.copy()
        params["addrs_to"] = [self.bob.address, slave.address]
        params["amounts"] = [100, 100]

        tx = TransferTokenTransaction.create(**params)
        tx.sign(self.alice)
        addresses_state = self.generate_addresses_state(tx)
        addresses_state[slave.address] = Mock(autospec=AddressState, name='slave AddressState',
                                              tokens={self.params["token_txhash"]: 0},
                                              transaction_hashes=[],
                                              balance=0)

        tx.apply_state_changes(addresses_state)

        self.assertEqual(addresses_state[self.alice.address].balance, 99)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)
        self.assertEqual(addresses_state[slave.address].balance, 0)
        addresses_state[self.alice.address].update_token_balance.assert_called_with(b'I declare the TEST token', -200)
        addresses_state[self.bob.address].update_token_balance.assert_called_with(b'I declare the TEST token', 100)
        addresses_state[slave.address].update_token_balance.assert_called_with(b'I declare the TEST token', 100)
        self.assertEqual([tx.txhash], addresses_state[self.alice.address].transaction_hashes)
        self.assertEqual([tx.txhash], addresses_state[self.bob.address].transaction_hashes)
        self.assertEqual([tx.txhash], addresses_state[slave.address].transaction_hashes)

        m_apply_state_PK.assert_called_once()

    def test_apply_state_changes_empty_addresses_state(self, m_logger, m_apply_state_PK, m_revert_state_PK):
        """
        Alice has 1000 tokens and 100 QRL, Bob has none. Alice sends some tokens to Bob.
        But this node has no AddressState corresponding to these parties.
        """
        tx = TransferTokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = {}
        tx.apply_state_changes(addresses_state)

        self.assertEqual({}, addresses_state)
        m_apply_state_PK.assert_called_once()

    def test_apply_state_changes_send_tokens_to_self(self, m_logger, m_apply_state_PK, m_revert_state_PK):
        """
        Alice has 1000 tokens and 100 QRL. She sends some tokens to herself. What happens next?
        """
        self.params["addrs_to"] = [self.alice.address]
        tx = TransferTokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = self.generate_addresses_state(tx)
        tx.apply_state_changes(addresses_state)

        self.assertEqual(addresses_state[self.alice.address].balance, 99)
        # Unfortunately importing mock.call results in some sort of ValueError so I can't check the arguments.
        self.assertEqual(addresses_state[self.alice.address].update_token_balance.call_count, 2)

        m_apply_state_PK.assert_called_once()

    def test_revert_state_changes(self, m_logger, m_apply_state_PK, m_revert_state_PK):
        """
        Alice has 1000 tokens and 100 QRL, Bob has none. Alice sends some tokens to Bob.
        Let's undo this.
        """
        tx = TransferTokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = self.generate_addresses_state(tx)
        addresses_state[self.alice.address].balance = 99
        addresses_state[self.alice.address].tokens[self.params["token_txhash"]] = 900
        addresses_state[self.alice.address].transaction_hashes = [tx.txhash]
        addresses_state[self.bob.address].balance = 0
        addresses_state[self.bob.address].tokens[self.params["token_txhash"]] = 100
        addresses_state[self.bob.address].transaction_hashes = [tx.txhash]

        tx.revert_state_changes(addresses_state, self.unused_chain_manager_mock)

        self.assertEqual(addresses_state[self.alice.address].balance, 100)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)
        addresses_state[self.alice.address].update_token_balance.assert_called_with(b'I declare the TEST token', 100)
        addresses_state[self.bob.address].update_token_balance.assert_called_with(b'I declare the TEST token', -100)
        self.assertEqual([], addresses_state[self.alice.address].transaction_hashes)
        self.assertEqual([], addresses_state[self.bob.address].transaction_hashes)

        m_revert_state_PK.assert_called_once()

    def test_revert_state_changes_multi_send(self, m_logger, m_apply_state_PK, m_revert_state_PK):
        """
        Alice has 1000 tokens and 100 QRL, Bob and Slave have none. Alice sends some tokens to Bob and Slave.
        Undo this.
        """
        slave = get_slave_xmss()
        self.params["addrs_to"] = [self.bob.address, slave.address]
        self.params["amounts"] = [100, 100]

        tx = TransferTokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = self.generate_addresses_state(tx)
        addresses_state[self.alice.address].balance = 99
        addresses_state[self.alice.address].tokens[self.params["token_txhash"]] = 900
        addresses_state[self.alice.address].transaction_hashes = [tx.txhash]
        addresses_state[self.bob.address].balance = 0
        addresses_state[self.bob.address].tokens[self.params["token_txhash"]] = 100
        addresses_state[self.bob.address].transaction_hashes = [tx.txhash]
        addresses_state[slave.address] = Mock(autospec=AddressState, name='slave AddressState',
                                              tokens={self.params["token_txhash"]: 100},
                                              transaction_hashes=[tx.txhash],
                                              balance=0)

        tx.revert_state_changes(addresses_state, self.unused_chain_manager_mock)

        self.assertEqual(addresses_state[self.alice.address].balance, 100)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)
        self.assertEqual(addresses_state[slave.address].balance, 0)
        addresses_state[self.alice.address].update_token_balance.assert_called_with(b'I declare the TEST token', 200)
        addresses_state[self.bob.address].update_token_balance.assert_called_with(b'I declare the TEST token', -100)
        addresses_state[slave.address].update_token_balance.assert_called_with(b'I declare the TEST token', -100)
        self.assertEqual([], addresses_state[self.alice.address].transaction_hashes)
        self.assertEqual([], addresses_state[self.bob.address].transaction_hashes)
        self.assertEqual([], addresses_state[slave.address].transaction_hashes)

        m_revert_state_PK.assert_called_once()

    def test_revert_state_changes_empty_addresses_state(self, m_logger, m_apply_state_PK, m_revert_state_PK):
        """
        If we didn't have any AddressStates for the addresses involved in this test, do nothing
        """
        tx = TransferTokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = {}

        tx.revert_state_changes(addresses_state, self.unused_chain_manager_mock)

        self.assertEqual(addresses_state, {})
        m_revert_state_PK.assert_called_once()

    def test_revert_state_changes_send_tokens_to_self(self, m_logger, m_apply_state_PK, m_revert_state_PK):
        """
        Alice has 1000 tokens and 100 QRL. She sends some tokens to herself.
        Can we undo this?
        """
        self.params["addrs_to"] = [self.alice.address]
        tx = TransferTokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = self.generate_addresses_state(tx)
        addresses_state[self.alice.address].balance = 99
        addresses_state[self.alice.address].transaction_hashes = [tx.txhash]
        tx.revert_state_changes(addresses_state, self.unused_chain_manager_mock)

        self.assertEqual(addresses_state[self.alice.address].balance, 100)
        # Unfortunately importing mock.call results in some sort of ValueError so I can't check the arguments.
        self.assertEqual(addresses_state[self.alice.address].update_token_balance.call_count, 2)

        m_revert_state_PK.assert_called_once()



from pyqrllib.pyqrllib import bin2hstr, QRLHelper

from qrl.core import config
from qrl.core.AddressState import AddressState
from qrl.core.misc import logger
from qrl.core.txs.Transaction import Transaction


class TokenTransaction(Transaction):
    """
    TokenTransaction to create new Token.
    """

    def __init__(self, protobuf_transaction=None):
        super(TokenTransaction, self).__init__(protobuf_transaction)

    @property
    def symbol(self):
        return self._data.token.symbol

    @property
    def name(self):
        return self._data.token.name

    @property
    def owner(self):
        return self._data.token.owner

    @property
    def decimals(self):
        return self._data.token.decimals

    @property
    def initial_balances(self):
        return self._data.token.initial_balances

    def get_data_bytes(self):
        data_bytes = (self.master_addr +
                      self.fee.to_bytes(8, byteorder='big', signed=False) +
                      self.symbol +
                      self.name +
                      self.owner +
                      self._data.token.decimals.to_bytes(8, byteorder='big', signed=False))

        for initial_balance in self._data.token.initial_balances:
            data_bytes += initial_balance.address
            data_bytes += initial_balance.amount.to_bytes(8, byteorder='big', signed=False)

        return data_bytes

    @staticmethod
    def create(symbol: bytes,
               name: bytes,
               owner: bytes,
               decimals: int,
               initial_balances: list,
               fee: int,
               xmss_pk: bytes,
               master_addr: bytes = None):
        transaction = TokenTransaction()

        if master_addr:
            transaction._data.master_addr = master_addr

        transaction._data.public_key = bytes(xmss_pk)

        transaction._data.token.symbol = symbol
        transaction._data.token.name = name
        transaction._data.token.owner = owner
        transaction._data.token.decimals = decimals

        for initial_balance in initial_balances:
            transaction._data.token.initial_balances.extend([initial_balance])

        transaction._data.fee = int(fee)

        transaction.validate_or_raise(verify_signature=False)

        return transaction

    def _validate_custom(self):
        if len(self.symbol) > config.dev.max_token_symbol_length:
            logger.warning('Token Symbol Length exceeds maximum limit')
            logger.warning('Found Symbol Length %s', len(self.symbol))
            logger.warning('Expected Symbol length %s', config.dev.max_token_symbol_length)
            return False

        if len(self.name) > config.dev.max_token_name_length:
            logger.warning('Token Name Length exceeds maximum limit')
            logger.warning('Found Name Length %s', len(self.symbol))
            logger.warning('Expected Name length %s', config.dev.max_token_name_length)
            return False

        if len(self.symbol) == 0:
            logger.warning('Missing Token Symbol')
            return False

        if len(self.name) == 0:
            logger.warning('Missing Token Name')
            return False

        if len(self.initial_balances) == 0:
            logger.warning('Invalid Token Transaction, without any initial balance')
            return False

        sum_of_initial_balances = 0
        for initial_balance in self.initial_balances:
            sum_of_initial_balances += initial_balance.amount
            if initial_balance.amount <= 0:
                logger.warning('Invalid Initial Amount in Token Transaction')
                logger.warning('Address %s | Amount %s', initial_balance.address, initial_balance.amount)
                return False

        allowed_decimals = self.calc_allowed_decimals(sum_of_initial_balances // 10 ** self.decimals)

        if self.decimals > allowed_decimals:
            logger.warning('Decimal is greater than maximum allowed decimal')
            logger.warning('Allowed Decimal %s', allowed_decimals)
            logger.warning('Decimals Found %s', self.decimals)
            return False

        if self.fee < 0:
            raise ValueError('TokenTransaction [%s] Invalid Fee = %d', bin2hstr(self.txhash), self.fee)

        return True

    # checks new tx validity based upon node statedb and node mempool.
    def validate_extended(self, addr_from_state: AddressState, addr_from_pk_state: AddressState):
        if not self.validate_slave(addr_from_state, addr_from_pk_state):
            return False

        tx_balance = addr_from_state.balance

        if not AddressState.address_is_valid(self.addr_from):
            logger.warning('Invalid address addr_from: %s', bin2hstr(self.addr_from))
            return False

        if not AddressState.address_is_valid(self.owner):
            logger.warning('Invalid address owner_addr: %s', bin2hstr(self.owner))
            return False

        for address_balance in self.initial_balances:
            if not AddressState.address_is_valid(address_balance.address):
                logger.warning('Invalid address in initial_balances: %s', bin2hstr(address_balance.address))
                return False

        if tx_balance < self.fee:
            logger.info('TokenTxn State validation failed for %s because: Insufficient funds', bin2hstr(self.txhash))
            logger.info('balance: %s, Fee: %s', tx_balance, self.fee)
            return False

        if addr_from_pk_state.ots_key_reuse(self.ots_key):
            logger.info('TokenTxn State validation failed for %s because: OTS Public key re-use detected',
                        bin2hstr(self.txhash))
            return False

        return True

    def apply_state_changes(self, addresses_state):
        addr_from_pk = bytes(QRLHelper.getAddress(self.PK))
        owner_processed = False
        addr_from_processed = False
        addr_from_pk_processed = False

        for initial_balance in self.initial_balances:
            if initial_balance.address == self.owner:
                owner_processed = True
            if initial_balance.address == self.addr_from:
                addr_from_processed = True
            if initial_balance.address == addr_from_pk:
                addr_from_pk_processed = True
            if initial_balance.address in addresses_state:
                addresses_state[initial_balance.address].update_token_balance(self.txhash, initial_balance.amount)
                addresses_state[initial_balance.address].transaction_hashes.append(self.txhash)

        if self.owner in addresses_state and not owner_processed:
            addresses_state[self.owner].transaction_hashes.append(self.txhash)

        if self.addr_from in addresses_state:
            addresses_state[self.addr_from].balance -= self.fee
            if not addr_from_processed and self.addr_from != self.owner:
                addresses_state[self.addr_from].transaction_hashes.append(self.txhash)

        if addr_from_pk in addresses_state:
            if self.addr_from != addr_from_pk and addr_from_pk != self.owner:
                if not addr_from_pk_processed:
                    addresses_state[addr_from_pk].transaction_hashes.append(self.txhash)
            addresses_state[addr_from_pk].increase_nonce()
            addresses_state[addr_from_pk].set_ots_key(self.ots_key)

    def revert_state_changes(self, addresses_state, chain_manager):
        addr_from_pk = bytes(QRLHelper.getAddress(self.PK))
        owner_processed = False
        addr_from_processed = False
        addr_from_pk_processed = False

        for initial_balance in self.initial_balances:
            if initial_balance.address == self.owner:
                owner_processed = True
            if initial_balance.address == self.addr_from:
                addr_from_processed = True
            if initial_balance.address == addr_from_pk:
                addr_from_pk_processed = True
            if initial_balance.address in addresses_state:
                addresses_state[initial_balance.address].update_token_balance(self.txhash,
                                                                              initial_balance.amount * -1)
                addresses_state[initial_balance.address].transaction_hashes.remove(self.txhash)

        if self.owner in addresses_state and not owner_processed:
            addresses_state[self.owner].transaction_hashes.remove(self.txhash)

        if self.addr_from in addresses_state:
            addresses_state[self.addr_from].balance += self.fee
            if not addr_from_processed and self.addr_from != self.owner:
                addresses_state[self.addr_from].transaction_hashes.remove(self.txhash)

        if addr_from_pk in addresses_state:
            if self.addr_from != addr_from_pk and addr_from_pk != self.owner:
                if not addr_from_pk_processed:
                    addresses_state[addr_from_pk].transaction_hashes.remove(self.txhash)
            addresses_state[addr_from_pk].decrease_nonce()
            addresses_state[addr_from_pk].unset_ots_key(self.ots_key, chain_manager)

    def set_affected_address(self, addresses_set: set):
        super().set_affected_address(addresses_set)
        addresses_set.add(self.owner)
        for initial_balance in self.initial_balances:
            addresses_set.add(initial_balance.address)
